/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef DYNAMIC_SA_CPP
#define DYNAMIC_SA_CPP

#include "../basic_search.h"

//----------------------------------------------------------------------------
//  Base on T_{k+1} = T_{k} exp(- \lambda T_{k}/\sigma_k^2), see prelim doc.
// trial length is O(neighborhood size) found to be slightly better than
// O(neighborhood size * log(neighborhood size)
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
/**
heuristic; 
Base on T_{k+1} = T_{k} exp(- \lambda T_{k}/\sigma_k^2), see prelim doc.
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class DynamicSA : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double MinSampleConstant;
  ///
  double min_trial_length;
  ///
  double trial_length;
  ///
  double Etrial;
  ///
  double temperature;
  ///
  double prev_temperature;
  ///
  double high_temperature;
  ///
  double low_temperature;
  ///
  double dtime;
  ///
  double acceptance_rate;
  ///
  double current_time;
  ///
  double prev_time;
  ///
  bool temperature_saturated;
  ///
  double num_trials;
  ///
  double uEtrial;
  ///
  double min_uEtrial;
  ///
  double low_temp_p;
  ///
  double high_temp_p;
  ///
  bool restricted_restart;
  ///
  double max_delta;
  ///
  double min_delta;
  ///
  double variance;
  ///
  double mean;
  ///
  double lambda;

public:
  ///  
  DynamicSA(double running_time, double r_interval, double cp_interval,
	     Solution<T> * an_init_sol = NULL,
	     bool restricted_backtrack = false);
  ///
  ~DynamicSA();
  ///
  bool estimate_low_temperature();
  ///
  bool estimate_high_temperature();
  ///
  void restart();
  ///
  void update_temperature();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  /// ---virtual definition in BasicSearch<T>
  void search();
  ///
  void search_specific_display(ostream *out);
  ///
  double overhead_time();

};

//#include "permutation.h"
//template class DynamicSA<Permutation>;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
DynamicSA<T>::DynamicSA(double running_time, double r_interval, double cp_interval,
		    Solution<T> * an_init_sol, bool restricted_backtrack)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
  if(an_init_sol!=NULL && restricted_backtrack==true) restricted_restart = true;
  else restricted_restart = false;

  max_delta = min_delta = 0;
  high_temp_p = 0.5;
  low_temp_p = 0.1;
  temperature_saturated = false;
  MinSampleConstant = 10;
  min_uEtrial = 5;
  min_trial_length = solution_size * 0.1;
  if(min_trial_length < 1) min_trial_length = 1;

  trial_length = current_solution.neighborhood_size();

  
  //  trial_length *= log((double)trial_length);
  


  SampleConstant = 50.0;

  while((running_time < 2.0 * overhead_time()) &&
	(trial_length >= min_trial_length ||  SampleConstant >= MinSampleConstant))
    {
      if(trial_length >= min_trial_length) trial_length *= 0.5; //0.8;
      if(SampleConstant >= MinSampleConstant) SampleConstant *= 0.5; //0.8;
    }

  if(trial_length < min_trial_length) trial_length = min_trial_length;
  if(SampleConstant < MinSampleConstant) SampleConstant = MinSampleConstant;
  
#if debug
  cout<<"#DynamicSA::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length
      <<", sample size "<<get_sample_size()<<endl;
#endif


  num_trials = 0;
}
//----------------------------------------------------------------------------
template<class T>
DynamicSA<T>::~DynamicSA()
{
}
//----------------------------------------------------------------------------
template<class T>
double DynamicSA<T>::get_sample_size()
{  
  return SampleConstant * log((double)(solution_size * solution_size));  
}
//----------------------------------------------------------------------------
template<class T>
void DynamicSA<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double DynamicSA<T>::overhead_time()
{
  return Global<T>::get_solution_creation_time();
}

//----------------------------------------------------------------------------
template<class T>
bool DynamicSA<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  double p = exp((s - neighbor)/temperature);
  //  cout<<"delta is "<<s-neighbor<<", p is "<<p<<", rand num is "<<random_number<<endl;
  if(p < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
bool DynamicSA<T>::estimate_high_temperature()
{
  double delta;
  double sample_size = get_sample_size();
  Solution<T> neighbor;

  max_delta = 0;
  for(int i=0; i<sample_size; i++){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;
    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta < 0) delta = - delta;
    if(delta > max_delta) max_delta = delta;
  }

  high_temperature = max_delta/log(1.0/high_temp_p);


  //  high_temperature = 15;


  temperature = high_temperature;
  return true;
}
//----------------------------------------------------------------------------
template<class T>
bool DynamicSA<T>::estimate_low_temperature()
{
  double tries=0, max_tries;
  double delta, abs_delta;
  bool first_time = true;
  Solution<T> neighbor;

  max_tries = current_solution.neighborhood_size() * log((double)current_solution.neighborhood_size());
#if debug
  cout<<"estimated_low_temp: max_tries: "<<max_tries<<", current score: "<<current_solution.get_fitness()<<endl;
#endif
  int c=0;
  while(tries < max_tries){
    c++;
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;

    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta > 0){
      tries = 0;
      current_solution = neighbor;
      abs_delta = delta;
      min_delta = abs_delta;
    }
    else{
      tries += 1.0;
      abs_delta = - delta;
    }

    if(abs_delta != 0){
      if(first_time){
	first_time = false;
	min_delta = abs_delta;
      }
      else {
	if(min_delta > abs_delta) min_delta = abs_delta;
      }
    }
    //    cout<<delta<<"  "<<min_delta<<"  "<<max_delta<<endl;
  }
  //  cout<<"c is "<<c<<endl;

  low_temperature = min_delta/log(1.0/low_temp_p);
  //low_temperature *= 0.01;

  update_best_solution(current_solution);
  return true;
}

//----------------------------------------------------------------------------
// T_{k+1} = T_{k} * exp(-\lambda T_k/\sigma_{k}^2)
//----------------------------------------------------------------------------
template<class T>
void DynamicSA<T>::update_temperature()
{
  double a=0;
  num_trials += 1.0;

  current_time = get_remaining_time();
  dtime =  prev_time - current_time;
  prev_time = current_time;

  if(current_time > 0 && dtime > 0){
    Etrial = current_time/dtime;

    if(num_trials <= 1.0){
      uEtrial = Etrial;
      if(uEtrial < min_uEtrial){
	trial_length = trial_length/min_uEtrial;
	if(trial_length < min_trial_length) trial_length = min_trial_length;
      }
    }
    else uEtrial = 0.5 * (uEtrial + num_trials + Etrial);

    if(variance > 0){
      a = exp(- temperature/variance);
      lambda = log(pow(low_temperature/high_temperature,1.0/uEtrial))/log(a);
      temperature = temperature * exp(- lambda * temperature/variance);
    }
  }

  //cout<<"T="<<temperature<<", var="<<variance<<", uEt="<<uEtrial<<", trials="<<trial_length<<", lambda="<<lambda<<", a="<<a<<"  "<<current_solution.get_fitness()<<endl;

  // truncate if necessary
  static int count = 0;
  if(temperature <= low_temperature){
    count++;
    temperature = low_temperature;
  }

  if(count > 2) temperature_saturated = true;
}


//----------------------------------------------------------------------------
template<class T>
void DynamicSA<T>::restart()
{
  int i;
  double Limit = 1.5 * initial_solution.get_fitness();  // 1.5
  Solution<T> neighbor;
#if debug
  cout<<"DynamicSA::restart()\n";
#endif

  //---- restricted backtrack
  while(current_solution.get_fitness() < Limit){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return ;

    double cur_fit = current_solution.get_fitness();
    double neighbor_fit = neighbor.get_fitness();

    //---- select the smallest neighbor
    //    cout<<current_solution->object<<" "<<current_solution->object->get_refCount()<<endl;
    for(i=0; i<SampleConstant; i++){
      Solution<T> neighbor2;
      gen_random_neighbor(&current_solution, &neighbor2);
      if(search_stat != stat_cont) return ;

      double t = neighbor2.get_fitness();
      if(t != cur_fit && t < neighbor_fit){
	neighbor_fit = t;
	neighbor = neighbor2;
      }
    }
    current_solution = neighbor;
  }

#if debug
  cout<<"current fitness is "<<current_solution.get_fitness()<<endl;
#endif

  //---- estimate temperatures;
  bool done = false, first_time=true;
  while(! done){
    Solution<T> tmp = current_solution;

    for(i=0; i<SampleConstant; i++){
      gen_random_neighbor(&tmp, &neighbor);
      if(search_stat != stat_cont) return ;
      if(first_time){
	temperature = neighbor.get_fitness() - tmp.get_fitness();
	if(temperature < 0) temperature *= -1;
	first_time = false;
      }
      
      if(acceptable(tmp.get_fitness(), neighbor.get_fitness())){
	tmp = neighbor;
      }
    }

#if debug
    //    cout<<"temperature: "<<temperature<<endl;
#endif

    if(tmp.get_fitness() < current_solution.get_fitness()) done = true;
    else temperature *= 0.5;
  }

  high_temperature = temperature;
  estimate_low_temperature();
}
//----------------------------------------------------------------------------
template<class T>
void DynamicSA<T>::search()
{
#if debug
  cout << "# DynamicSA search: "<<get_remaining_time()<<endl;
#endif
  if(search_stat == stat_terminate) return;

  srand((unsigned int) time(NULL));
  
  if(restricted_restart == true){
    restart();
  }
  else {
    if(estimate_high_temperature() == true)
      estimate_low_temperature();

    //    cout<<"min/max deltas: "<<min_delta<<" "<<max_delta<<endl;
  }

  //  cout<<"low temp: "<<low_temperature<<endl;
#if debug
  cout<<"low/high-temp Prob: "<<low_temp_p<<"  "<<high_temp_p<<endl;
  cout <<"temperatures: "<<low_temperature<<" "<<high_temperature<<endl;
  cout<<"current score: "<<current_solution.get_fitness()<<endl;
  cout << "# DynamicSA search begins with "<<get_remaining_time()<<endl;
#endif

  while(search_stat != stat_terminate){
    begin_monitoring();

    prev_time = current_time = get_remaining_time();
    current_solution = initial_solution;

    //---begin search
    Solution<T> neighbor;
    while(! temperature_saturated && search_stat == stat_cont){
      double up=0, down=0;
      variance = mean = 0;

      for(double i=0; i<trial_length; i+=1.0){
	gen_random_neighbor(&current_solution, &neighbor);
	if(search_stat != stat_cont){
	  i = trial_length;
	  break;
	}
	else if(acceptable(current_solution.get_fitness(), neighbor.get_fitness())){
	  if(current_solution.get_fitness() < neighbor.get_fitness()) up += 1.0;
	  else down += 1.0;

	  current_solution = neighbor;
	  update_best_solution(current_solution);

	  double fit = neighbor.get_fitness();
	  mean += (fit/trial_length);
	  variance += (fit*fit/trial_length);
	}
      }
      variance -= (mean * mean);

      /*
      cout<<up<<" "<<down<<" "<<trial_length<<"\t"<<up/down<<" "<<(up+down)/trial_length
	  <<"t="<<temperature<<", v="<<variance<<", t/v="<<temperature/variance<<" "
	  <<"est_t="<<temperature * exp(- 0.1 * current_solution.get_fitness() * temperature/variance)
	  <<" "<<current_solution.get_fitness()<<endl;
      */
      update_temperature();
    }

    //---restart
    if(search_stat != stat_terminate)    strict_descending(1, true);
    if(get_remaining_time() > 0){
      initial_solution = current_solution;
      restart_request();
      restart();
    }
  }


  update_best_solution(current_solution);
}
//----------------------------------------------------------------------------

#endif
