/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef GCF_SA_CPP
#define GCF_SA_CPP

#include "../basic_search.h"
/*
code: g_new_sa (power 3, neighborhood size n log(n), adaptive trial length, greedy selection)

Algorithm:  annealing according to an given np-rate (rate of acceptance 
of negative vs positive moves)

NP-Rate:
--------
np-rate: exponential decay -- T total time, t_r remaining time.  Then,
   np-rate(t_r) = exp(t_r * log(2)/T) - 1;
This satisfies: (1) beginning np-rate: 1, (2) final np-rate: 0
The exponential constant is e.

Temperature:
-----------
given np-rate r (between 0 and 1), find temperature t so that the np-rate
is close to r.
 
beginning temperature: estimate_temperature(1)
temperature schedule: 

Algorithm:
---------
set beginning temperature for r=1;
run sa at this temperture;
get_rate at current remaining time;
lower temperature to that rate;

*/


//----------------------------------------------------------------------------
/**
heuristic; 
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class GCF_SA : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double GreedyConstant;
  ///
  double max_trial_length;
  ///
  double trial_length;
  ///
  double min_uEtrial;
  ///
  double temperature;
  ///
  double total_time;
  ///
  double rate;

public:
  
  ///
  GCF_SA(double running_time, double r_interval, double cp_interval,
	 Solution<T> * an_init_sol = NULL,
	 bool restricted_backtrack = false);
  
  ///
  ~GCF_SA();

  ///
  double get_rate();
  ///
  void update_temperature(double ups, double downs, double up_fit);
  ///
  bool estimate_initial_temperature();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  /// ---virtual definition in BasicSearch<T>
  void search();
  ///
  void search_specific_display(ostream *out);
  ///
  double overhead_time();

};

//#include "permutation.h"
//template class GCF_SA<Permutation>;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
GCF_SA<T>::GCF_SA(double running_time, double r_interval, double cp_interval,
		    Solution<T> * an_init_sol, bool restricted_backtrack)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
  GreedyConstant = 10.0;
  total_time = get_remaining_time();
  trial_length = current_solution.neighborhood_size();
  // S*log(S) seems to be the right number //
  trial_length *= log((double)trial_length);
  
  min_uEtrial = 10;
  // assuming that GreedyConstant neighbors are created per trial
  max_trial_length = get_remaining_time() / (GreedyConstant * min_uEtrial * Global<T>::get_neighbor_creation_time());
  if(max_trial_length < 1) max_trial_length = 1;
  if(trial_length > max_trial_length){
    GreedyConstant *= 0.5;
    max_trial_length = get_remaining_time() / (GreedyConstant * min_uEtrial * Global<T>::get_neighbor_creation_time());
    if(max_trial_length < 1) max_trial_length = 1;
    trial_length = max_trial_length;
  }

  SampleConstant = 50.0;
  //  while(get_sample_size() > 0.5 * trial_length) SampleConstant *= 0.9;
  while(get_sample_size() > 0.1 * trial_length) SampleConstant *= 0.9;

#if debug
  cout<<"#GCF_SA::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length
      <<", sample size "<<get_sample_size()<<endl;
#endif

}
//----------------------------------------------------------------------------
template<class T>
GCF_SA<T>::~GCF_SA()
{
}
//----------------------------------------------------------------------------
template<class T>
double GCF_SA<T>::get_sample_size()
{  
  return SampleConstant * log((double)(solution_size * solution_size));  
}
//----------------------------------------------------------------------------
template<class T>
void GCF_SA<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double GCF_SA<T>::overhead_time()
{
  return Global<T>::get_solution_creation_time();
}

//----------------------------------------------------------------------------
template<class T>
bool GCF_SA<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  double p = exp((s - neighbor)/temperature);
  //  cout<<"delta is "<<s-neighbor<<", p is "<<p<<", rand num is "<<random_number<<endl;
  if(p < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
void GCF_SA<T>::update_temperature(double ups, double downs, double up_fit)
{
  /*
    find T so that (sum exp(up_fits/T))/downs = rate
    short cut:  exp(mean_fit) * ups/downs = rate
  */
  double mean_up_fit = up_fit/ups;
  rate = get_rate();
  if(ups > 0 && downs > 0){
    temperature = mean_up_fit/log(rate * downs / ups);
  }

#if debug
  cout<<rate<<"  "<<ups/downs<<" "<<(ups+downs)/trial_length<<"\t"<<ups<<" "<<downs<<" "<<trial_length<<"\t"<<temperature
      <<"\t"<< current_solution.get_fitness()<<endl;

#endif

}
//----------------------------------------------------------------------------
template<class T>
bool GCF_SA<T>::estimate_initial_temperature()
{
  double delta;
  double sample_size = get_sample_size();
  double ups=0, downs=0, up_fit=0;
  Solution<T> neighbor;

  for(int i=0; i<sample_size; i++){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;
    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta < 0){
      ups += 1.0;
      up_fit += delta;
    }
    else {
      downs += 1.0;
    }
  }

  update_temperature(ups, downs, up_fit);
  return true;
}

//----------------------------------------------------------------------------
template<class T>
double GCF_SA<T>::get_rate()
{
  /*
    exponential rate:   np-rate(t_r) = exp(t_r * log(1+r_0) / T) - 1;
    This satisfies: (1) beginning np-rate: starting_rate, (2) final np-rate: 0
    The exponential constant is e.
  */
  const double starting_rate = 0.75;
  const double exp_const = 3.0;
  static bool first_time = true;
  if(first_time){
    total_time = get_remaining_time();
    first_time = false;
  }
  double t_r = get_remaining_time();
  return  exp(log(1.0 + starting_rate) * pow(t_r/total_time, exp_const)) - 1.0;
}

//----------------------------------------------------------------------------
template<class T>
void GCF_SA<T>::search()
{
#if debug
  cout << "# GCF_SA search: "<<get_remaining_time()<<endl;
#endif
  if(search_stat == stat_terminate) return;

  srand((unsigned int) time(NULL));

  if(estimate_initial_temperature() == true){
    begin_monitoring();

    //    cout<<"temp: "<<temperature<<endl;

    current_solution = initial_solution;

    //---begin search
    Solution<T> neighbor, best_neighbor;
    while(search_stat == stat_cont){
      double ups = 0, downs = 0, up_fit = 0;
      for(double i=0; i<trial_length; i+=1.0){
	for(int j=0; j<GreedyConstant; j++){
	  gen_random_neighbor(&current_solution, &neighbor);
	  if(search_stat != stat_cont) break;
	  if(j==0) best_neighbor = neighbor;
	  else if(neighbor.get_fitness() <= best_neighbor.get_fitness())
	    best_neighbor = neighbor;
	}

	if(search_stat != stat_cont) break;
	else if(acceptable(current_solution.get_fitness(), best_neighbor.get_fitness())){
	  double delta = current_solution.get_fitness() - best_neighbor.get_fitness();
	  if(delta < 0){
	    ups += 1.0;
	    up_fit += delta;
	  }
	  else downs += 1.0;

	  current_solution = best_neighbor;
	  update_best_solution(current_solution);
	}

	if(search_stat != stat_cont) break;


	//	gen_random_neighbor(&current_solution, &neighbor);
	//	if(search_stat != stat_cont){
	//	  i = trial_length;
	//	  break;
	//	}
	//	else if(acceptable(current_solution.get_fitness(), neighbor.get_fitness())){
	//	  double delta = current_solution.get_fitness() - neighbor.get_fitness();
	//	  if(delta < 0){
	//	    ups += 1.0;
	//	    up_fit += delta;
	//	  }
	//	  else downs += 1.0;
	//	  current_solution = neighbor;
	//	  update_best_solution(current_solution);
	//	}
      }
      
      update_temperature(ups, downs, up_fit);
    }
  }

  //  cout<<"temp: "<<temperature<<endl;
  if(search_stat != stat_terminate)    strict_descending(1, true);

  update_best_solution(current_solution);
}
//----------------------------------------------------------------------------

#endif
