/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef G_SMART_SA_CPP
#define G_SMART_SA_CPP

#include "../basic_search.h"
#include <math.h>

//----------------------------------------------------------------------------
// This version implements the basic "exponential" simulated annealing heuristic
//
// idea:  get parameters such as Max temp, Min temp, by doing a hill climb ***
//
// get rid of artificial acceptance rate?????  simplify all these.
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
/**
heuristic; 
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class GSmartSA : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double MinSampleConstant;
  ///
  double max_trial_length;
  ///
  double trial_length;
  ///
  double Etrial;
  ///
  double temperature;
  ///
  double prev_temperature;
  ///
  double high_temperature;
  ///
  double low_temperature;
  ///
  double dtime;
  ///
  double dtemp;
  ///
  double acceptance_rate;
  ///
  double current_time;
  ///
  double prev_time;
  ///
  double min_uEtrial;
  ///
  bool temperature_saturated;

  ///
  double num_trials;
  ///
  double uEtrial;
  ///
  double low_temp_p;
  ///
  double high_temp_p;
  ///
  bool restricted_restart;

  ///
  double GreedyConstant;

public:
  
  ///
  GSmartSA(double running_time, double r_interval, double cp_interval,
	  Solution<T> * an_init_sol = NULL,
	  bool restricted_backtrack = false);
  
  ///
  ~GSmartSA();

  ///
  bool estimate_low_temperature();
  ///
  bool estimate_high_temperature();
  ///
  void restart();
  ///
  void update_temperature();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  /// ---virtual definition in BasicSearch<T>
  void search();

  ///
  void search_specific_display(ostream *out);

  ///
  double overhead_time();

};

//#include "permutation.h"
//template class GSmartSA<Permutation>;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
GSmartSA<T>::GSmartSA(double running_time, double r_interval, double cp_interval,
		    Solution<T> * an_init_sol, bool restricted_backtrack)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
	//  GreedyConstant = 10;
  GreedyConstant = 5;

  if(an_init_sol!=NULL && restricted_backtrack==true) restricted_restart = true;
  else restricted_restart = false;

  high_temp_p = 0.5;
  low_temp_p = 0.1;
  temperature_saturated = false;
  //  MinSampleConstant = 10;
  min_uEtrial = 10;

  trial_length = current_solution.neighborhood_size();
  trial_length *= log((double)trial_length);

  max_trial_length = get_remaining_time() / (min_uEtrial * Global<T>::get_neighbor_creation_time());
  if(trial_length > max_trial_length) trial_length = max_trial_length;

  SampleConstant = 50.0;
  while(get_sample_size() > 0.1 * trial_length) SampleConstant *= 0.9;

#if debug
  cout<<"#GSmartSA::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length
      <<", sample size "<<get_sample_size()<<endl;
#endif


  num_trials = 0;
}
//----------------------------------------------------------------------------
template<class T>
GSmartSA<T>::~GSmartSA()
{
}
//----------------------------------------------------------------------------
template<class T>
double GSmartSA<T>::get_sample_size()
{  
  return SampleConstant * log((double)(solution_size * solution_size));  
}
//----------------------------------------------------------------------------
template<class T>
void GSmartSA<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double GSmartSA<T>::overhead_time()
{
  return Global<T>::get_solution_creation_time();
}

//----------------------------------------------------------------------------
template<class T>
bool GSmartSA<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  double p = exp((s - neighbor)/temperature);
  //  cout<<"delta is "<<s-neighbor<<", p is "<<p<<", rand num is "<<random_number<<endl;
  if(p < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
bool GSmartSA<T>::estimate_high_temperature()
{
  double max_delta, delta;
  double sample_size = get_sample_size();
  Solution<T> neighbor;

  max_delta = 0;
  for(int i=0; i<sample_size; i++){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;
    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta < 0) delta = - delta;
    if(delta > max_delta) max_delta = delta;
  }

  high_temperature = max_delta/log(1.0/high_temp_p);


  //  high_temperature = 15;


  temperature = high_temperature;
  return true;
}
//----------------------------------------------------------------------------
template<class T>
bool GSmartSA<T>::estimate_low_temperature()
{
  double tries=0, max_tries;
  double min_delta, delta, abs_delta;
  bool first_time = true;
  Solution<T> neighbor;

  max_tries = current_solution.neighborhood_size() * log((double)current_solution.neighborhood_size());
#if debug
  cout<<"estimated_low_temp: max_tries: "<<max_tries<<", current score: "<<current_solution.get_fitness()<<endl;
#endif
  while(tries < max_tries){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;

    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta > 0){
      tries = 0;
      current_solution = neighbor;
      abs_delta = delta;
      min_delta = abs_delta;
    }
    else{
      tries += 1.0;
      abs_delta = - delta;
    }

    if(abs_delta != 0){
      if(first_time){
	first_time = false;
	min_delta = abs_delta;
      }
      else {
	if(min_delta > abs_delta) min_delta = abs_delta;
      }
    }
    //    cout<<delta<<"  "<<min_delta<<"  "<<max_delta<<endl;
  }

  low_temperature = min_delta/log(1.0/low_temp_p);
  //low_temperature *= 0.01;

  return true;
}

//----------------------------------------------------------------------------
template<class T>
void GSmartSA<T>::update_temperature()
{
  num_trials += 1.0;

  current_time = get_remaining_time();
  dtime =  prev_time - current_time;
  prev_time = current_time;

  //**** This seems to work for graph coloring
  if(current_time > 0 && dtime > 0){
    Etrial = current_time/dtime;

    if(num_trials <= 1.0){
      uEtrial = Etrial;
      //      if(uEtrial < min_uEtrial){
      //	trial_length = trial_length/min_uEtrial;
      //	if(trial_length < min_trial_length) trial_length = min_trial_length;
      //      }
    }
    else uEtrial = 0.5 * (uEtrial + num_trials + Etrial);
    dtemp = pow(low_temperature/high_temperature, 1.0/uEtrial);
    if(dtemp > 0.9999999) dtemp = 0.9999999;
    if(dtemp < 0.0001) dtemp = 0.0001;
  }

  temperature *= dtemp;

  //  cout<<"temp="<<temperature<<", dtemp="<<dtemp<<", Et="<<Etrial<<", uEt="<<uEtrial<<", num_trials="<<num_trials<<endl;

  // truncate if necessary
  static int count = 0;
  if(temperature <= low_temperature){
    count++;
    temperature = low_temperature;
  }

  if(count > 2) temperature_saturated = true;
}


//----------------------------------------------------------------------------
template<class T>
void GSmartSA<T>::restart()
{
  int i;
  double Limit = 1.5 * initial_solution.get_fitness();  // 1.5
  Solution<T> neighbor;

  //---- restricted backtrack
  while(current_solution.get_fitness() < Limit){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return ;

    double cur_fit = current_solution.get_fitness();
    double neighbor_fit = neighbor.get_fitness();

    //---- select the smallest neighbor
    //    cout<<current_solution->object<<" "<<current_solution->object->get_refCount()<<endl;
    for(i=0; i<SampleConstant; i++){
      Solution<T> neighbor2;
      gen_random_neighbor(&current_solution, &neighbor2);
      if(search_stat != stat_cont) return ;

      double t = neighbor2.get_fitness();
      if(t != cur_fit && t < neighbor_fit){
	neighbor_fit = t;
	neighbor = neighbor2;
      }
    }
    current_solution = neighbor;
  }

#if debug
  cout<<"current fitness is "<<current_solution.get_fitness()<<endl;
#endif

  //---- estimate temperatures;
  bool done = false, first_time=true;
  while(! done){
    Solution<T> tmp = current_solution;

    for(i=0; i<SampleConstant; i++){
      gen_random_neighbor(&tmp, &neighbor);
      if(search_stat != stat_cont) return ;
      if(first_time){
	temperature = neighbor.get_fitness() - tmp.get_fitness();
	if(temperature < 0) temperature *= -1;
	first_time = false;
      }
      
      if(acceptable(tmp.get_fitness(), neighbor.get_fitness())){
	tmp = neighbor;
      }
    }

#if debug
    //    cout<<"temperature: "<<temperature<<endl;
#endif

    if(tmp.get_fitness() < current_solution.get_fitness()) done = true;
    else temperature *= 0.5;
  }

  high_temperature = temperature;
  estimate_low_temperature();
}
//----------------------------------------------------------------------------
template<class T>
void GSmartSA<T>::search()
{
#if debug
  cout << "# GSmartSA search: "<<get_remaining_time()<<endl;
#endif
  if(search_stat == stat_terminate) return;

  srand((unsigned int) time(NULL));
  
  if(restricted_restart == true){
    restart();
  }
  else {
    if(estimate_high_temperature() == true)
      estimate_low_temperature();
  }

#if debug
  cout<<"low/high-temp Prob: "<<low_temp_p<<"  "<<high_temp_p<<endl;
  cout <<"temperatures: "<<low_temperature<<" "<<high_temperature<<endl;
  cout<<"current score: "<<current_solution.get_fitness()<<endl;
  cout << "# GSmartSA search begins with "<<get_remaining_time()<<endl;
#endif

  while(search_stat != stat_terminate){
    begin_monitoring();

    if(low_temperature >= high_temperature) high_temperature = low_temperature * 2;

    prev_time = current_time = get_remaining_time();
    current_solution = initial_solution;

    //---begin search
    Solution<T> neighbor, best_neighbor;
    while(! temperature_saturated && search_stat == stat_cont){
      for(double i=0; i<trial_length; i+=1.0){
	for(int j=0; j<GreedyConstant; j++){
	  gen_random_neighbor(&current_solution, &neighbor);
	  if(search_stat != stat_cont) break;
	  if(j==0) best_neighbor = neighbor;
	  else if(neighbor.get_fitness() <= best_neighbor.get_fitness())
	    best_neighbor = neighbor;
	}

	if(search_stat != stat_cont) break;
	else if(acceptable(current_solution.get_fitness(), best_neighbor.get_fitness())){
	  current_solution = best_neighbor;
	  update_best_solution(current_solution);
	}

	if(search_stat != stat_cont) break;
      }
      
      update_temperature();
    }

    //---restart
    if(search_stat != stat_terminate)    strict_descending(1, true);
    if(get_remaining_time() > 0){
      initial_solution = current_solution;
      restart_request();
      restart();
    }

  }


  update_best_solution(current_solution);
}
//----------------------------------------------------------------------------

#endif
