/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef RATE_SA_CPP
#define RATE_SA_CPP

#include "../basic_search.h"

//----------------------------------------------------------------------------
// This version implements the basic "exponential" simulated annealing heuristic
//
// idea:  get parameters such as Max temp, Min temp, by doing a hill climb ***
//
// get rid of artificial acceptance rate?????  simplify all these.
//
// ratesa is different from smart_sa in that low temperature is selected so
// that the probability acceptance at low temperature is satisfied.
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
/**
heuristic
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class RateSA : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double MinSampleConstant;
  ///
  double min_trial_length;
  ///
  double trial_length;
  ///
  double Etrial;
  ///
  double temperature;
  ///
  double prev_temperature;
  ///
  double high_temperature;
  ///
  double low_temperature;
  ///
  double dtime;
  ///
  double dtemp;
  ///
  double acceptance_rate;
  ///
  double current_time;
  ///
  double prev_time;
  ///
  bool temperature_saturated;

  ///
  double num_trials;
  ///
  double uEtrial;
  ///
  double min_uEtrial;
  ///
  double low_temp_p;
  ///
  double high_temp_p;
  ///
  bool restricted_restart;

  ///
  double max_delta, min_delta;

public:
  
  ///
  RateSA(double running_time, double r_interval, double cp_interval,
	 Solution<T> * an_init_sol = NULL,
	 bool restricted_backtrack = false);
  
  ///
  ~RateSA();

  ///
  bool estimate_low_temperature();
  ///
  bool estimate_high_temperature();
  ///
  void estimate_restart_temperature();
  ///
  void update_temperature();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  ///
  bool anneal(double N, double *ups, double *downs);

  /// ---virtual definition in BasicSearch<T>
  void search();

  ///
  void search_specific_display(ostream *out);

  ///
  double overhead_time();

};

//#include "permutation.h"
//template class RateSA<Permutation>;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
RateSA<T>::RateSA(double running_time, double r_interval, double cp_interval,
		  Solution<T> * an_init_sol, bool restricted_backtrack)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
  if(an_init_sol!=NULL && restricted_backtrack==true) restricted_restart = true;
  else restricted_restart = false;

  max_delta = min_delta = 0;
  high_temp_p = 0.5;
  low_temp_p = 0.1;
  temperature_saturated = false;
  MinSampleConstant = 10;
  min_uEtrial = 5;
  min_trial_length = solution_size * 0.1;
  if(min_trial_length < 1) min_trial_length = 1;

  trial_length = current_solution.neighborhood_size();
  trial_length *= log((double)trial_length);
  
  SampleConstant = 50.0;

  while((running_time < 2.0 * overhead_time()) &&
	(trial_length >= min_trial_length ||  SampleConstant >= MinSampleConstant))
    {
      if(trial_length >= min_trial_length) trial_length *= 0.5; //0.8;
      if(SampleConstant >= MinSampleConstant) SampleConstant *= 0.5; //0.8;
    }

  if(trial_length < min_trial_length) trial_length = min_trial_length;
  if(SampleConstant < MinSampleConstant) SampleConstant = MinSampleConstant;
  
#if debug
  cout<<"#RateSA::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length
      <<", sample size "<<get_sample_size()<<endl;
#endif


  num_trials = 0;
}
//----------------------------------------------------------------------------
template<class T>
RateSA<T>::~RateSA()
{
}
//----------------------------------------------------------------------------
template<class T>
double RateSA<T>::get_sample_size()
{  
  return SampleConstant * log((double)(solution_size * solution_size));  
}
//----------------------------------------------------------------------------
template<class T>
void RateSA<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double RateSA<T>::overhead_time()
{
  return Global<T>::get_solution_creation_time();
}

//----------------------------------------------------------------------------
template<class T>
bool RateSA<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  double p = exp((s - neighbor)/temperature);
  //  cout<<"delta is "<<s-neighbor<<", p is "<<p<<", rand num is "<<random_number<<endl;
  if(p < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
bool RateSA<T>::anneal(double N, double *ups, double *downs)
{
  Solution<T> neighbor;
  *ups = *downs = 0;
  for(int i=0; i<N; i++){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;
    else if(acceptable(current_solution.get_fitness(), neighbor.get_fitness())){
      if(current_solution.get_fitness() < neighbor.get_fitness()) *ups += 1.0;
      else *downs += 1.0;
      
      current_solution = neighbor;
      update_best_solution(current_solution);
    }
  }

  //  cout<<*ups<<" "<<*downs<<" "<<N<<"\t"<<*ups / *downs<<" "<<(*ups+*downs)/N<<"  "<<temperature<<endl;

  return true;
}
//----------------------------------------------------------------------------
template<class T>
bool RateSA<T>::estimate_high_temperature()
{
  double delta;
  double sample_size = get_sample_size();
  Solution<T> neighbor;

  max_delta = 0;
  for(int i=0; i<sample_size; i++){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;
    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta < 0) delta = - delta;
    if(delta > max_delta) max_delta = delta;
  }

  high_temperature = max_delta/log(1.0/high_temp_p);
  return true;
}
//----------------------------------------------------------------------------
template<class T>
bool RateSA<T>::estimate_low_temperature()
{
  double tries=0, max_tries;
  double delta, abs_delta;
  bool first_time = true;
  Solution<T> neighbor;

  max_tries = current_solution.neighborhood_size() * log((double)current_solution.neighborhood_size());
#if debug
  cout<<"estimated_low_temp: max_tries: "<<max_tries<<", current score: "<<current_solution.get_fitness()<<endl;
#endif
  int c=0;
  while(tries < max_tries){
    c++;
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return false;

    delta = current_solution.get_fitness() - neighbor.get_fitness();
    if(delta > 0){
      tries = 0;
      current_solution = neighbor;
      abs_delta = delta;
      min_delta = abs_delta;
    }
    else{
      tries += 1.0;
      abs_delta = - delta;
    }

    if(abs_delta != 0){
      if(first_time){
	first_time = false;
	min_delta = abs_delta;
      }
      else {
	if(min_delta > abs_delta) min_delta = abs_delta;
      }
    }
    //    cout<<delta<<"  "<<min_delta<<"  "<<max_delta<<endl;
  }
  //  cout<<"c is "<<c<<endl;

  // ---- Now select low temperature at the right rates

  low_temperature = min_delta/log(1.0/low_temp_p);
  double ups, downs;
  double N = get_sample_size();
  bool stop = false;
  while(! stop){
    temperature = low_temperature;
    
    anneal(N, &ups, &downs);

#if 0
    cout<<"sampling rates... "<<temperature<<" "<<ups<<" "<<downs<<" "<<N<<"\t"<<ups/downs<<" "<<(ups+downs)/N<<"\n";
#endif

    if(ups == 0 || (ups/downs < 0.01 && (ups+downs)/N < 0.01))
      stop = true;
    else low_temperature *= 0.5;
  }

  //  cout<<"low temp: "<<low_temperature<<endl;
  update_best_solution(current_solution);
  return true;
}

//----------------------------------------------------------------------------
template<class T>
void RateSA<T>::update_temperature()
{
  num_trials += 1.0;

  current_time = get_remaining_time();
  dtime =  prev_time - current_time;
  prev_time = current_time;

  if(current_time > 0 && dtime > 0){
    Etrial = current_time/dtime;

    if(num_trials <= 1.0){
      uEtrial = Etrial;
      if(uEtrial < min_uEtrial){
	trial_length = trial_length/min_uEtrial;
	if(trial_length < min_trial_length) trial_length = min_trial_length;
      }
    }
    else uEtrial = 0.5 * (uEtrial + num_trials + Etrial);
    dtemp = pow(low_temperature/high_temperature, 1.0/uEtrial);
    if(dtemp > 0.9999999) dtemp = 0.9999999;
    if(dtemp < 0.0001) dtemp = 0.0001;
  }

  temperature *= dtemp;

#if 0
  cout<<"temp="<<temperature<<", dtemp="<<dtemp<<", Et="<<Etrial<<", uEt="<<uEtrial<<", num_trials="<<num_trials<<"\t"<<trial_length<<"\t"<<current_solution.get_fitness()<<endl;
#endif

  // truncate if necessary
  static int count = 0;
  if(temperature <= low_temperature){
    count++;
    temperature = low_temperature;
  }

  if(count > 2) temperature_saturated = true;
}


//----------------------------------------------------------------------------
template<class T>
void RateSA<T>::estimate_restart_temperature()
{
  int i;
  double Limit = 1.5 * initial_solution.get_fitness();  // 1.5
  Solution<T> neighbor;
#if debug
  cout<<"RateSA::estimate_restart_temperature()\n";
#endif

  //---- restricted backtrack
  while(current_solution.get_fitness() < Limit){
    gen_random_neighbor(&current_solution, &neighbor);
    if(search_stat != stat_cont) return ;

    double cur_fit = current_solution.get_fitness();
    double neighbor_fit = neighbor.get_fitness();

    //---- select the smallest neighbor
    //    cout<<current_solution->object<<" "<<current_solution->object->get_refCount()<<endl;
    for(i=0; i<SampleConstant; i++){
      Solution<T> neighbor2;
      gen_random_neighbor(&current_solution, &neighbor2);
      if(search_stat != stat_cont) return ;

      double t = neighbor2.get_fitness();
      if(t != cur_fit && t < neighbor_fit){
	neighbor_fit = t;
	neighbor = neighbor2;
      }
    }
    current_solution = neighbor;
  }

#if debug
  cout<<"current fitness is "<<current_solution.get_fitness()<<endl;
#endif

  //---- estimate temperatures;
  bool done = false, first_time=true;
  while(! done){
    Solution<T> tmp = current_solution;

    for(i=0; i<SampleConstant; i++){
      gen_random_neighbor(&tmp, &neighbor);
      if(search_stat != stat_cont) return ;
      if(first_time){
	temperature = neighbor.get_fitness() - tmp.get_fitness();
	if(temperature < 0) temperature *= -1;
	first_time = false;
      }
      
      if(acceptable(tmp.get_fitness(), neighbor.get_fitness())){
	tmp = neighbor;
      }
    }

    if(tmp.get_fitness() < current_solution.get_fitness()) done = true;
    else temperature *= 0.5;
  }

  high_temperature = temperature;
  estimate_low_temperature();
}
//----------------------------------------------------------------------------
template<class T>
void RateSA<T>::search()
{
#if debug
  cout << "# RateSA search: "<<get_remaining_time()<<endl;
#endif
  if(search_stat == stat_terminate) return;

  srand((unsigned int) time(NULL));
  
  if(restricted_restart == true){
    estimate_restart_temperature();
  }
  else {
    if(estimate_high_temperature() == true)
      estimate_low_temperature();
  }

#if debug
  cout<<"low/high-temp Prob: "<<low_temp_p<<"  "<<high_temp_p<<endl;
  cout <<"temperatures: "<<low_temperature<<" "<<high_temperature<<endl;
  cout<<"current score: "<<current_solution.get_fitness()<<endl;
  cout << "# RateSA search begins with "<<get_remaining_time()<<endl;
#endif

  temperature = high_temperature;
  while(search_stat != stat_terminate){
    begin_monitoring();

    prev_time = current_time = get_remaining_time();
    current_solution = initial_solution;

    //---begin search
    Solution<T> neighbor;
    while(! temperature_saturated && search_stat == stat_cont){
      double up=0, down=0;
      if(anneal(trial_length, &up, &down) == false) break;
      update_temperature();
    }

    //---restart
    if(search_stat != stat_terminate)    strict_descending(1, true);
    if(get_remaining_time() > 0){
      initial_solution = current_solution;
      restart_request();
      estimate_restart_temperature();
    }
  }


  update_best_solution(current_solution);
}
//----------------------------------------------------------------------------

#endif
