/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef SA_RESTART_CPP
#define SA_RESTART_CPP

#include "../basic_search.h"

//----------------------------------------------------------------------------
// This version implements the basic "exponential" simulated annealing heuristic
//
// idea:  get parameters such as Max temp, Min temp, by doing a hill climb ***
//
// get rid of artificial acceptance rate?????  simplify all these.
//----------------------------------------------------------------------------

struct temperature_record {
  double temperature;
  int count, count2;
};

//----------------------------------------------------------------------------
/**
heuristic
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class SARestart : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double MinSampleConstant;
  ///
  double min_trial_length;
  ///
  double trial_length;
  ///
  double expected_trials;
  ///
  double temperature;
  ///
  double prev_temperature;
  ///
  double high_temperature;
  ///
  double low_temperature;
  ///
  double dtime;
  ///
  double dtemp;
  ///
  double acceptance_rate;
  ///
  double current_time;
  ///
  double prev_time;
  ///
  bool temperature_saturated;
  ///
  bool must_exit;
  ///
  int DownwardMoveConst;
  ///
  int low_temperature_count;

  /// ----variables for restarting
  temperature_record   restart_point;
  ///
  static const double critical_prob;
  ///
  double first_search_time;
  ///
  double local_get_remaining_time();
  ///
  bool search_loop(bool must_exit);
  ///
  void prerestart_setup();

public:
  
  ///
  SARestart(double running_time, double r_interval, double cp_interval,
	    Solution<T> * an_init_sol = NULL);
  
  ///
  ~SARestart();

  ///
  void presearch_setup();
  ///
  void update_temperature();
  ///
  void estimate_maximum_temperature();
  ///
  void estimate_minimum_temperature();
  ///
  void strict_descending();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  /// ---virtual definition in BasicSearch<T>
  void search();

  ///
  void search_specific_display(ostream *out);

  ///
  double overhead_time();

  ///
  void search_exit();
};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

template<class T>
const double SARestart<T>::critical_prob = 0.5;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
SARestart<T>::SARestart(double running_time, double r_interval, double cp_interval,
			Solution<T> * an_init_sol)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
  DownwardMoveConst = 20;

  temperature_saturated = false;
  MinSampleConstant = 10;
  min_trial_length = solution_size * 0.1;

  //  double Trial_Length_Constant = 0.5;
  //  trial_length = Trial_Length_Constant * solution_size * solution_size;
  //  SampleConstant = 150.0;

  trial_length = current_solution->neighborhood_size();
  SampleConstant = 100.0;


  while((running_time < 2.0 * overhead_time()) &&
	(trial_length >= min_trial_length ||  SampleConstant >= MinSampleConstant))
    {
      if(trial_length >= min_trial_length) trial_length *= 0.5; //0.8;
      if(SampleConstant >= MinSampleConstant) SampleConstant *= 0.5; //0.8;
    }

  if(trial_length < min_trial_length) trial_length = min_trial_length;
  if(SampleConstant < MinSampleConstant) SampleConstant = MinSampleConstant;
  
#if debug
  cout<<"#SARestart::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length<<endl;
#endif

  //  MinLength = solution_size * solution_size;
  //  trial_length = solution_size;
  //  SampleConstant = 150.0;
}
//----------------------------------------------------------------------------
template<class T>
SARestart<T>::~SARestart()
{
}
//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::search_exit()
{
#if debug
  cout<<"high, low, marked "<<high_temperature<<" "<<low_temperature<<" "<<restart_point.temperature<<endl; 
#endif
}
//----------------------------------------------------------------------------
template<class T>
double SARestart<T>::get_sample_size()
{
  return SampleConstant * log(solution_size * solution_size);
}
//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double SARestart<T>::overhead_time()
{
  double min_time =
    // presearch
    get_sample_size() * 3.0 * Global<T>::get_solution_creation_time()
    // at least two trials
    + 2.0 * trial_length * (Global<T>::get_neighbor_creation_time() + 0.5*Global<T>::get_solution_creation_time());

#if debug
  cout<<"#SARestart::overhead_time() minimum time "<<min_time<<endl;
  cout<<"#SARestart::overhead_time() trial_length "<< trial_length
      <<", neighbor_time "<<Global<T>::get_neighbor_creation_time()
      <<", solution_time "<<Global<T>::get_solution_creation_time()<<endl;
#endif
  return min_time;
}

//----------------------------------------------------------------------------
template<class T>
bool SARestart<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  if(exp((s - neighbor)/temperature) < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::strict_descending()
{
  double tries=0, max_tries;
#if debug
  cout<<"#SARestart::strict_descending()\n";
#endif

  max_tries =  10 * current_solution->neighborhood_size() * log(current_solution->neighborhood_size());
  //solution_size*solution_size*log(solution_size*solution_size);

  while(tries < max_tries){
    Solution<T> *neighbor=NULL;

    neighbor = gen_random_neighbor(current_solution);

    if(neighbor == NULL) break;
    else if(current_solution->get_fitness() > neighbor->get_fitness()){
      tries = 0;
      delete current_solution;
      current_solution = neighbor;
    }
    else{
      delete neighbor;
      tries += 1.0;
    }
  }
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::estimate_minimum_temperature()
{
  Solution<T> *s1=NULL, *s2=NULL;
  double delta_fit, max_delta, min_delta, sample_size, i;

  sample_size = get_sample_size();

  for(i=0; i<sample_size; i+=1.0){
    s1 = gen_random_solution();
    if(s1 == NULL) break;

    s2 = gen_random_solution();
    if(s2 == NULL) break;

    delta_fit = s1->get_fitness() - s2->get_fitness();
    if(delta_fit < 0) delta_fit *= -1.0;
    if(i < 1.0) max_delta = min_delta = delta_fit;
    else {
      if(delta_fit > max_delta) max_delta = delta_fit;
      if(delta_fit < min_delta) min_delta = delta_fit;
    }
    delete s1;
    delete s2;
  }

  // 99% within largest/smallest temperature
  low_temperature = min_delta/(7.0*log(solution_size * solution_size));

  //********
  low_temperature = low_temperature * 0.01;
  if(low_temperature < 0.000001) low_temperature = 0.000001;
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::estimate_maximum_temperature()
{
  Solution<T> *s=NULL;
  double fit, max_fit, min_fit, sample_size, i;

  sample_size = get_sample_size();

  for(i=0; i<sample_size; i+=1.0){
    s = gen_random_solution();
    if(s == NULL) break;

    fit = s->get_fitness();
    if(i < 1.0) max_fit = min_fit = fit;
    else {
      if(fit > max_fit) max_fit = fit;
      if(fit < min_fit) min_fit = fit;
    }
    delete s;
  }
  high_temperature = (max_fit - min_fit)/log(1.0/acceptance_rate);
  //**************
  high_temperature = high_temperature * 0.01;
}

//----------------------------------------------------------------------------
template<class T>
double SARestart<T>::local_get_remaining_time()
{
  return get_remaining_time() - first_search_time;
}
//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::update_temperature()
{
  current_time = local_get_remaining_time();
  dtime =  prev_time - current_time;
  prev_time = current_time;

  if(current_time > 0 && dtime > 0){
    expected_trials = current_time/dtime;
    dtemp = pow(low_temperature/temperature, 1.0/expected_trials);
    if(dtemp > 0.9999999) dtemp = 0.9999999;
    if(dtemp < 0.0001) dtemp = 0.0001;
  }

  temperature *= dtemp;

  //  cout<<temperature<<" "<<dtemp<<" "<<current_time<<" "<<dtime<<" "<<expected_trials<<endl;

  // truncate if necessary
  if(temperature <= low_temperature){
    low_temperature_count++;
    temperature = low_temperature;
  }

  if(low_temperature_count > 2) temperature_saturated = true;
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::presearch_setup()
{
  srand(time(NULL));

  low_temperature_count = 0;
  temperature_saturated = false;
  must_exit = false;

  // ------------------------------------------------------------
  // assume best case O(n^2) acceptance rate, O(n^3) start length
  // 0 < acceptance_rate < 1
  // reminder: log(x) < 0 if x<1.   log(x) undefined if x < 0.
  // ------------------------------------------------------------
  double minimum_time = overhead_time();
  if(minimum_time < 1.0) minimum_time = 1.00001;
  acceptance_rate = log(minimum_time)/ (2.0*log(solution_size*solution_size) + 
					log(Global<T>::get_neighbor_creation_time()));
  
  if(acceptance_rate > 0.9) acceptance_rate = 0.9;
  if(acceptance_rate < 0.1) acceptance_rate = 0.1;
  
  //---- Now estimate lowest and highest temperature
  estimate_minimum_temperature();

  estimate_maximum_temperature();
  temperature = high_temperature;

  //----- Restart setup
  restart_point.temperature = temperature;
  restart_point.count = restart_point.count2 = 0;

  first_search_time = critical_prob * get_remaining_time();
  prev_time = current_time = local_get_remaining_time();
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::prerestart_setup()
{
  restart_request();

  low_temperature_count = 0;
  temperature_saturated = false;
  must_exit = false;

  temperature = restart_point.temperature;
  restart_point.count = restart_point.count2 = 0;

  first_search_time = 0;
  prev_time = current_time = local_get_remaining_time();

#if debug
  cout<<"#restart at temperature "<<temperature<<"\t(h,l)  "<<high_temperature<<" "<<low_temperature<<endl;
  cout.flush();
#endif
}
//----------------------------------------------------------------------------
template<class T>
bool SARestart<T>::search_loop(bool must_exit)
{
  Solution<T> *neighbor = NULL;
  for(double i=0; i<trial_length; i+=1.0){
    (restart_point.count)++;

    neighbor = gen_random_neighbor(current_solution);

    if(neighbor == NULL){
#if debug
      cout<<"neighbor is NULL, i = "<<i<<endl; cout.flush();
#endif
      i = trial_length;
      must_exit = true;
    }
    else if(acceptable(current_solution->get_fitness(), neighbor->get_fitness())){
      delete current_solution;
      current_solution = neighbor;
      
      update_best_solution(*current_solution);
      
      //--- gather restart statistics
      if(restart_point.count <= 1.0/critical_prob){ // downward moves in the last 1/p tries

	(restart_point.count2)++; // number of downward moves
	if(restart_point.count2 >= DownwardMoveConst){
	  restart_point.temperature = temperature;
	  restart_point.count2 = 0;
	}
      }
      else restart_point.count2 = 0;

      restart_point.count = 0;
    }
    else{
      delete neighbor;
    }
  }
  update_temperature();

  return must_exit;
}

//----------------------------------------------------------------------------
template<class T>
void SARestart<T>::search()
{
#if debug
  cout << "# SARestart search begins\n";
#endif

  presearch_setup();
  begin_monitoring();

#if debug
  cout <<"#after pre-search: "<<high_temperature<<" "<<low_temperature<<" "<<prev_time<<endl;
#endif

  //---begin search
  while(! temperature_saturated && !must_exit){
    must_exit = search_loop(must_exit);
  }

  //---restart search
#if debug
  cout <<"#before restart\n";  cout.flush();
#endif

  prerestart_setup();
  while(! temperature_saturated && !must_exit){
    must_exit = search_loop(must_exit);
  }

#if debug
  cout <<"#after restart "<<must_exit<<" "<<temperature_saturated<<endl;  cout.flush();
#endif

  if(! must_exit) strict_descending();

  update_best_solution(*current_solution);
}
//----------------------------------------------------------------------------

#endif
