/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef SIMULATED_ANNEAL_CPP
#define SIMULATED_ANNEAL_CPP

#include "../basic_search.h"

//----------------------------------------------------------------------------
// This version implements the basic "exponential" simulated annealing heuristic
//
// idea:  get parameters such as Max temp, Min temp, by doing a hill climb ***
//
// get rid of artificial acceptance rate?????  simplify all these.
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
/**
heuristic
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class SimulatedAnneal : public BasicSearch<T> {
private:
  ///
  double SampleConstant;
  ///
  double MinSampleConstant;
  ///
  double min_trial_length;
  ///
  double trial_length;
  ///
  double Etrial;
  ///
  double temperature;
  ///
  double prev_temperature;
  ///
  double high_temperature;
  ///
  double low_temperature;
  ///
  double dtime;
  ///
  double dtemp;
  ///
  double acceptance_rate;
  ///
  double current_time;
  ///
  double prev_time;
  ///
  bool temperature_saturated;
  ///
  double num_trials;
  ///
  double uEtrial;
  ///
  double LOW_TEMP_MULT;

public:
  
  ///
  SimulatedAnneal(double running_time, double r_interval, double cp_interval,
		  Solution<T> * an_init_sol = NULL);
  
  ///
  ~SimulatedAnneal();

  ///
  void presearch_setup();
  ///
  void update_temperature();
  ///
  void estimate_maximum_temperature();
  ///
  void estimate_minimum_temperature();
  ///
  bool acceptable(double s, double neighbor);
  ///
  double get_sample_size();

  /// ---virtual definition in BasicSearch<T>
  void search();

  ///
  void search_specific_display(ostream *out);

  ///
  double overhead_time();

};

//#include "permutation.h"
//template class SimulatedAnneal<Permutation>;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
// IMPLEMENTATIONS
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
template<class T>
SimulatedAnneal<T>::SimulatedAnneal(double running_time, double r_interval, double cp_interval,
				    Solution<T> * an_init_sol)
  : BasicSearch<T> (running_time, r_interval, cp_interval, an_init_sol)
{
  LOW_TEMP_MULT = 0.01;
  temperature_saturated = false;
  MinSampleConstant = 10;
  min_trial_length = solution_size * 0.1;

  //  double Trial_Length_Constant = 0.5;
  //  trial_length = Trial_Length_Constant * solution_size * solution_size;
  //  SampleConstant = 150.0;

  trial_length = current_solution.neighborhood_size();
  SampleConstant = 100.0;

  while((running_time < 2.0 * overhead_time()) &&
	(trial_length >= min_trial_length ||  SampleConstant >= MinSampleConstant))
    {
      if(trial_length >= min_trial_length) trial_length *= 0.5; //0.8;
      if(SampleConstant >= MinSampleConstant) SampleConstant *= 0.5; //0.8;
    }

  if(trial_length < min_trial_length) trial_length = min_trial_length;
  if(SampleConstant < MinSampleConstant) SampleConstant = MinSampleConstant;
  
#if debug
  cout<<"#SimulatedAnneal::constructor() SampleConstant "<<SampleConstant<<", trial_length "<<trial_length
      <<", sample size "<<get_sample_size()<<endl;
#endif


  num_trials = 0;
}
//----------------------------------------------------------------------------
template<class T>
SimulatedAnneal<T>::~SimulatedAnneal()
{
}
//----------------------------------------------------------------------------
template<class T>
double SimulatedAnneal<T>::get_sample_size()
{  
  return SampleConstant * log((double)(solution_size * solution_size));  
}
//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::search_specific_display(ostream *out)
{
  *out << "(temperature, "<< temperature << ")\n";
}
//----------------------------------------------------------------------------
template<class T>
double SimulatedAnneal<T>::overhead_time()
{
  double min_time =
    // presearch
    get_sample_size() * 3.0 * Global<T>::get_solution_creation_time()
    // at least two trials
    + 2.0 * trial_length * (Global<T>::get_neighbor_creation_time() + 0.5*Global<T>::get_solution_creation_time());

#if debug
  cout<<"#SimulatedAnneal::overhead_time() minimum time "<<min_time<<endl;
  cout<<"#SimulatedAnneal::overhead_time() trial_length "<< trial_length
      <<", neighbor_time "<<Global<T>::get_neighbor_creation_time()
      <<", solution_time "<<Global<T>::get_solution_creation_time()<<endl;
#endif
  return min_time;
}

//----------------------------------------------------------------------------
template<class T>
bool SimulatedAnneal<T>::acceptable(double s, double neighbor)
{
  if(neighbor <= s) return true;
  double random_number = (double)rand() / (double)RAND_MAX;
  if(exp((s - neighbor)/temperature) < random_number) return false;
  return true;
}

//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::estimate_minimum_temperature()
{
  Solution<T> s1,s2;
  double delta_fit, max_delta, min_delta, sample_size, i;

  sample_size = get_sample_size();

  for(i=0; i<sample_size; i+=1.0){
    gen_random_solution(&s1);
    if(search_stat != stat_cont) break;
    gen_random_solution(&s2);
    if(search_stat != stat_cont) break;

    delta_fit = s1.get_fitness() - s2.get_fitness();
    if(delta_fit < 0) delta_fit *= -1.0;
    if(i < 1.0) max_delta = min_delta = delta_fit;
    else {
      if(delta_fit > max_delta) max_delta = delta_fit;
      if(delta_fit < min_delta) min_delta = delta_fit;
    }
  }

  // 99% within largest/smallest temperature
  low_temperature = min_delta/(7.0*log((double)(solution_size * solution_size)));
  low_temperature = low_temperature * LOW_TEMP_MULT;

  if(low_temperature < 0.000001) low_temperature = 0.000001;
}

//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::estimate_maximum_temperature()
{
  Solution<T> s;
  double fit, max_fit, min_fit, sample_size, i;

  sample_size = get_sample_size();

  for(i=0; i<sample_size; i+=1.0){
     gen_random_solution(&s);
     if(search_stat != stat_cont) break;

    fit = s.get_fitness();
    if(i < 1.0) max_fit = min_fit = fit;
    else {
      if(fit > max_fit) max_fit = fit;
      if(fit < min_fit) min_fit = fit;
    }
  }
  high_temperature = (max_fit - min_fit)/log(1.0/acceptance_rate);

  //*********
  //  high_temperature = high_temperature * 0.01;

  temperature = high_temperature;
}

//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::update_temperature()
{
  num_trials += 1.0;

  current_time = get_remaining_time();
  dtime =  prev_time - current_time;
  prev_time = current_time;

  //**** This seems to work for graph coloring
  if(current_time > 0 && dtime > 0){
    Etrial = current_time/dtime;

    if(num_trials <= 1.0) uEtrial = Etrial;
    else uEtrial = 0.5 * (uEtrial + num_trials + Etrial);

#if 0
    dtemp = pow(low_temperature/temperature, 1.0/Etrial);
#else
    dtemp = pow(low_temperature/high_temperature, 1.0/uEtrial);
#endif
    if(dtemp > 0.9999999) dtemp = 0.9999999;
    if(dtemp < 0.0001) dtemp = 0.0001;

    //    trial_length = (uEtrial+trial_length)*0.5;
    //    if(trial_length < min_trial_length) trial_length = min_trial_length;  
  }

  temperature *= dtemp;

  //  cout<<"temp="<<temperature<<", dtemp="<<dtemp<<", Et="<<Etrial<<", uEt="<<uEtrial<<", num_trials="<<num_trials<<endl;

  // truncate if necessary
  static int count = 0;
  if(temperature <= low_temperature){
    count++;
    temperature = low_temperature;
  }

  if(count > 2) temperature_saturated = true;
}

//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::presearch_setup()
{
  srand((unsigned int) time(NULL));

  // ------------------------------------------------------------
  // assume best case O(n^2) acceptance rate, O(n^3) start length
  // 0 < acceptance_rate < 1
  // reminder: log(x) < 0 if x<1.   log(x) undefined if x < 0.
  // ------------------------------------------------------------
  double minimum_time = overhead_time();
  if(minimum_time < 1.0) minimum_time = 1.00001;
  acceptance_rate = log(minimum_time)/ (2.0*log((double)(solution_size*solution_size)) + 
					log(Global<T>::get_neighbor_creation_time()));
  
  //  cout<<"acceptance_rate: "<<acceptance_rate<<endl;
  if(acceptance_rate > 0.9) acceptance_rate = 0.9;
  if(acceptance_rate < 0.1) acceptance_rate = 0.1;
  //  cout<<"acceptance_rate: "<<acceptance_rate<<endl;
  
  //---- Now estimate lowest and highest temperature
  estimate_minimum_temperature();

  estimate_maximum_temperature();

  prev_time = current_time = get_remaining_time();
}
//----------------------------------------------------------------------------
template<class T>
void SimulatedAnneal<T>::search()
{
#if debug
  cout << "# SimulatedAnneal search begins: "<<get_remaining_time()<<endl;
#endif

  if(search_stat == stat_terminate) return;

  presearch_setup();
  begin_monitoring();

#if debug
  cout <<"#after pre-search: "<<high_temperature<<" "<<low_temperature<<" "<<prev_time<<endl;
  cout<<"\t\tsearch stat: "<<search_stat<<" "<<temperature_saturated<<" "<<trial_length<<endl;
#endif
  //---begin search
  Solution<T> neighbor;

  while(! temperature_saturated && search_stat == stat_cont){
    //    double a = Global<T>::time_left();
    for(double i=0; i<trial_length; i+=1.0){
      gen_random_neighbor(&current_solution, &neighbor);

      if(search_stat != stat_cont){
	i = trial_length;
	break;
      }
      else if(acceptable(current_solution.get_fitness(), neighbor.get_fitness())){
	current_solution = neighbor;
	update_best_solution(current_solution);
      }

      //      cout<<current_solution.object<<" "<<current_solution.get_fitness()<<"\t"
      //	  <<neighbor.object<<" "<<neighbor.get_fitness()<<endl;
    }
    //    double b = Global<T>::time_left();
    //    cout<<"\t"<<trial_length<<" taking "<<(a-b)*1000<<endl;

    update_temperature();
  }

  if(search_stat != stat_terminate)    strict_descending(10, true);

  update_best_solution(current_solution);
}
//----------------------------------------------------------------------------

#endif
