/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

This software uses the Graph Template Library (GTL) copyrighted by the
University of Passau, http://www.fmi.uni-passau.de/Graphlet/GTL.
*/

#ifndef TSP_GRAPHS_H
#define TSP_GRAPHS_H

#include <GTL/graph.h>
#include <GTL/dfs.h>
#include <GTL/debug.h>
#include <cstdlib>
#include <cassert>
#include <fstream>
#include <iostream>
#include <vector.h>

#include "../errors.h"

#ifndef NoSuchEdge
#define NoSuchEdge 1000000
#endif

//--------------------------------------------------------------------
//---- EDGE HANDLER
// In order to correctly assign weights to edges, we maintain a structure.
// This structure is populated after the parsing of every edge.
// It is read, and an edge map is produced, once the graph has been wholly read.
//--------------------------------------------------------------------

typedef struct EdgeHandler{
	edge e;
	double w;
} edge_handler;


//--------------------------------------------------------------------
//---- WEIGHTED GRAPH
// The adjacency matrix weight graph class extends the GTL graph class
// in order to take advantage of weight information.
//--------------------------------------------------------------------

/**
@author Vinhthuy Phan
@version 1.0
*/
class weighted_graph : public graph{

private:
  /// An intermediate structure to contain edge, weight pairs.
  std::vector<edge_handler> weight_vector;

  /// used to map nodes to their index.
  node_map<int> node_index_map;

  /// used to map indexes to nodes. The first node index is 1.
  node *indexed_nodes;

  /// The adjacency matrix
  float **adjacency_matrix;

  ///
  int node_i;

public:
  weighted_graph() : graph(){node_i = 1;}
  ~weighted_graph();

  ///  An implementation of the base class' virtual function. Loads edge, weight
  ///  pairs into weight_vector.
  void load_edge_info_handler (edge e, GML_pair* list);

  ///  An implementation of the base class' virtual function. Is
  ///  called once the entire graph has been loaded. It uses the graph information
  ///  to fill in the maps and indices.
  void load_graph_info_handler(GML_pair* list);

  /// Given an edge (x,y) two node indices, returns the edge weight.
  float edge_weight(int x, int y);

  /// Given an edge (x,y) two node indices, returns existence.
  bool edge_exists(int x, int y);

  ///	
  int index_from_node(node & x);

  ///	
  node & node_from_index(int index);

  ///
  void load_node_info_handler (node v, GML_pair* list);
};

/// --------------------------------------------------------------------
/// ---- GRAPH USED FOR TSP
/// --------------------------------------------------------------------
/**
@author Vinhthuy Phan
@version 1.0
*/
class Graph 
{
public:
  /// GTL weighted graph
  weighted_graph G;   

  ///
  int num_of_nodes;
  ///
  int num_of_edges;

  ///
  Graph(char *GML_file_name);
  ~Graph();

  /// report the number of nodes.
  static int size;
  ///
  static int get_size();
  ///
  int get_num_of_edges();
  ///
  node & node_from_index(int index);
  ///
  int index_from_node(node & x);
  ///
  void display();
  ///
  void display(const char *out_file);
  ///
  double edge_weight(int x, int y);

  /// return the existence of an edge given by the two nodes it is adjacent to.
  bool edge_exists(int x, int y);
};


/// --------------------------------------------------------------------

#endif
