/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

This software uses the Graph Template Library (GTL) copyrighted by the
University of Passau, http://www.fmi.uni-passau.de/Graphlet/GTL.
*/
#include "vc_graph.h"

//--------------------------------------------------------------------
// DIMACS Format
//--------------------------------------------------------------------

int UndirectedGraph::size = 0;

int UndirectedGraph::get_size()
{
  return size;
}

UndirectedGraph::UndirectedGraph(char *file_name)
{
   edges = vertices = 0;
  ifstream in(file_name);
  if(!in){
    cerr << "can't open "<<file_name<<endl;
    exit(0);
  }
  string line;
  istringstream *sline;

  for (getline(in, line); !in.eof(); getline(in, line)){
    if(line[0] == 'c'){
      if(line == "c directed graph"){
        cout << "this is supposed to be undirected graph.\n";
        exit(0);
      }
    }
    else if(line[0] == 'p'){
      string problem_description, problem_token;
      sline = new istringstream(line);
      *sline >> problem_token >> problem_description >> vertices >> edges;

      vertex_weight = new int [vertices + 1];
      for(int j=1; j<=vertices; j++) vertex_weight[j] = 1;
      neighbor = new edge_set [vertices + 1];
      delete sline;
    }
    else if(line[0] == 'n'){
      sline = new istringstream(line);
      string node_token;
      int node_id;
      *sline >> node_token >> node_id;
      *sline >> vertex_weight[node_id];
      delete sline;
    }
    else if(line[0] == 'e'){
      string edge_token;
      sline = new istringstream(line);
      int source, dest;
      double weight;
      Edge forward_e, backward_e;
      *sline >> edge_token >> source >> dest >> weight;
      forward_e.vertex = dest;
      backward_e.vertex = source;
      forward_e.weight = backward_e.weight = weight;
      neighbor[source].insert(forward_e);
      neighbor[dest].insert(backward_e);
      delete sline;
    }
    else {
      cerr << "UndirectedGraph::UndirectedGraph: unknown type\n";
      exit(0);
    }
  }

  in.close();
  size = vertices;
}

//--------------------------------------------------------------------
UndirectedGraph::~UndirectedGraph()
{
  if(vertices > 0) {
    delete [] neighbor;
    delete [] vertex_weight;
  }
}
//--------------------------------------------------------------------

int UndirectedGraph::get_vertex_weight(int index) const
{
  if(index < 1 || index > vertices) {
    cerr << "Illegal access to node "<<index<<endl;
    exit(0);
  }
  return vertex_weight[index];
}

//--------------------------------------------------------------------
bool UndirectedGraph::edge_exist(int source, int dest) const
{
  if(source<1 || source>vertices) non_exist_error(source, "edge_exist()");
  if(dest<1 || dest>vertices) non_exist_error(dest, "edge_exist()");
  Edge e;
  e.vertex = dest;
  edge_set::iterator it = neighbor[source].find(e);
  return (it != neighbor[source].end());
}
//--------------------------------------------------------------------
bool UndirectedGraph::get_edge_weight(int source, int dest, double *weight) const
{
  if(source < 1 || source>vertices || dest<1 || dest>vertices)
    {
      cerr << "Illegal access to edge "<<source<<", "<<dest<<endl;
      exit(0);
    }
  Edge e;
  e.vertex = dest;
  edge_set::iterator it = neighbor[source].find(e);
  if(it == neighbor[source].end()){
    *weight = 0;
    return false;
  }
  *weight = it->weight;
  return true;
}

//--------------------------------------------------------------------
edge_set * UndirectedGraph::get_edges(int v)
{
  if(v<1 || v>vertices) non_exist_error(v, "get_edges()");
  return &(neighbor[v]);
}
//--------------------------------------------------------------------
int UndirectedGraph::degree(int v) const
{
  if(v<1 || v>vertices) non_exist_error(v, "degree()");
  return neighbor[v].size();
}
//--------------------------------------------------------------------
ostream & operator<<(ostream & out, UndirectedGraph & g)
{
  int i;
  out << g.vertices <<" vertices, "<<g.edges<< " edges\n";
  for(i=1; i<=g.vertices; i++){
    out << "edge "<< i << " (w="
        <<g.get_vertex_weight(i)<<", d="<<g.degree(i)<<") --> ";
    edge_set::iterator e = g.neighbor[i].begin();
    for( ; e!=g.neighbor[i].end(); e++){
      out << "[" << e->vertex << " ("<<e->weight<<")] ";
    }
    out << endl;
  }
  return out;
}
//--------------------------------------------------------------------
bool UndirectedGraph::remove_edge(int source, int dest) 
{
  if(source<1 || source>vertices) non_exist_error(source, "remove_edge()");
  if(dest<1 || dest>vertices) non_exist_error(dest, "remove_edge()");
  Edge e1, e2;
  e1.vertex = dest;
  e2.vertex = source;
  edge_set::iterator it1 = neighbor[source].find(e1), it2 = neighbor[dest].find(e2);
  if(it1 == neighbor[source].end() || it2 == neighbor[dest].end()) return false;
  neighbor[source].erase(*it1);
  neighbor[dest].erase(*it2);
  edges--;
  return true;
}
//--------------------------------------------------------------------
bool UndirectedGraph::remove_edges_containing(int v)
{
  if(v<1 || v>vertices) non_exist_error(v, "remove_edges_containing()");
  edge_set::iterator it, begin=neighbor[v].begin(), end=neighbor[v].end();
  for(it=begin; it!=end; it++){
    Edge e;
    e.vertex = v;
    neighbor[it->vertex].erase(e);
  }
  neighbor[v].erase(neighbor[v].begin(), neighbor[v].end());
  return true;
}
//--------------------------------------------------------------------
