/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/

#ifndef BASIC_SEARCH_H
#define BASIC_SEARCH_H

#include "basic_types.h"
#include "monitor.h"
#include "solution.h"
#include "objective_function.h"
#include "global.h"

/**
the base class of all search heuristics.
@author Vinhthuy Phan
@version 1.0
*/
template<class T>
class BasicSearch {

private:
  ///
  double running_time;
  ///
  double report_interval;
  ///
  double check_progress_interval;
  ///
  double MaxProgInterval;
  ///
  double MinProgInterval;
  ///
  Monitor *monitor;
  ///
  void reset_progress_window();
  ///
  void compute_solution_times(double running_time);
  ///
  bool monitor_flag, no_progress;
  ///
  bool check_progress_now();

  /** This method must be invoked to reset solution-generation block (if possible)
       when solution cannot be generated (blocked by the assigned monitor). */
  Status check_monitor_status();

  ///
  bool final_report;


protected:
  ///
  const static int progress_window = 20;

  /** based on Chebychev inequality
      var/mean <= epsilon/16 ==> P(X within +/- epsilon/2 of mean) >= 3/4
      epsilon = 0.04 ==> 98% */
//  const static double variance_threshold = 0.0016/16.0;
  static double variance_threshold;

  ///
  int solution_size;
  ///
  Status search_stat;
  ///
  double sampled_fit[progress_window];
  ///
  double sum_fit;
  ///
  double sum_squares_fit;
  ///
  bool sample_is_full;
  int window_position;

  /** These methods and only them are used exclusively by derived search
      heuristics to obtain new solutions  */
  void gen_random_solution(Solution<T> * new_sol, int n = -1);
  ///
  void gen_random_neighbor(Solution<T> * s, Solution<T> * new_sol);
  ///
  void update_best_solution(Solution<T> &s);
  ///
  void reset_window();
  ///
  void prepare_to_terminate();
  ///
  void clean_up();
  ///
  void restart_request();
  ///
  double cp_count;
public:
  
  /** Each search must make sure that the best solution throughout the search
      is stored at *best_solution when it finishes.*/
  Solution<T> initial_solution;
  ///
  Solution<T> current_solution;
  ///
  Solution<T> best_solution;
  
  ///
  BasicSearch(double running_time, double report_interval, double check_progress_interval,
	      Solution<T> * an_init_sol = NULL);
  
  ///
  void strict_descending(double sample_constant, bool forced);

  ///
  void initialize(Monitor *mon, Solution<T> * init_sol, ostream *out);
  
  ///
  void report_now();

  ///
  bool is_enough_time();

  ///
  double get_remaining_time();

  ///
  void begin_monitoring() { monitor_flag = true; }

  ///
  bool should_monitoring() { return monitor_flag; }

  ///
  void set_final_report() { final_report = true; }

  ///
  void reset_final_report() { final_report = false; }

  //---- Pure Virtual Functions
  ///
  virtual ~BasicSearch() ;
  
  ///
  virtual void search() = 0;
  
  ///
  virtual void search_specific_display(ostream *out) = 0;

  ///
  virtual double overhead_time() = 0;

};

template <class T> double BasicSearch<T>::variance_threshold = 0.0016/16.0;

//#include "permutation.h"
//template class BasicSearch<Permutation>;
//#include "basic_search.h"

//----------------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------------
template<class T>
BasicSearch<T>::BasicSearch(double running_time,
			    double report_interval,
			    double check_progress_interval,
			    Solution<T> * an_init_sol)
{
#if 0//debug
  cout<<"#BasicSearch::BasicSearch(): running_time "<<running_time
      <<", report_int "<<report_interval<<", check_prog_int "<<check_progress_interval<<endl;
#endif

  cp_count = 0;
  no_progress = final_report = false;
  monitor_flag = false;
  this->running_time = running_time;
  this->report_interval = report_interval;
  this->check_progress_interval = check_progress_interval;
  MinProgInterval = 0.05;
  MaxProgInterval = check_progress_interval;
  search_stat = stat_cont;

  monitor = new Monitor(search_mode,
			running_time,
			report_interval,
			check_progress_interval,
			NULL);
  //  cout<<"RR: "<<report_interval<<" "<<check_progress_interval<<endl;

  solution_size = ObjectiveFunction::get_size();
  if(an_init_sol == NULL) {
    gen_random_solution(&initial_solution, solution_size);
    if(search_stat == stat_terminate){
      return;
    }
  }
  //  else initial_solution.copy(*an_init_sol);
  else initial_solution = *an_init_sol;

  current_solution =  initial_solution;
  best_solution = initial_solution;
  reset_window();
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::clean_up()
{
  ;
}
//----------------------------------------------------------------------------------
template<class T>
BasicSearch<T>::~BasicSearch()
{
  //  cout<<"cp_count is "<<cp_count<<endl;
  clean_up();
}

//----------------------------------------------------------------------------------

template<class T>
double BasicSearch<T>::get_remaining_time() { return monitor->time_left(); }

//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::update_best_solution(Solution<T> &s)
{
  if(best_solution.object == NULL || best_solution.get_fitness() > s.get_fitness()){
    //    best_solution.copy(s);
    //    best_solution.consolidate();
    best_solution = s;
  }
}
//----------------------------------------------------------------------------------
template<class T>
bool BasicSearch<T>::is_enough_time()
{
  if(overhead_time() > monitor->time_left()){
#if debug
    cout <<"#not enough time,  overhead time: "<<overhead_time()
	 <<", time left "<<monitor->time_left()<<endl;
#endif
    return false;
  }
  return true;
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::reset_window()
{
  sum_fit = sum_squares_fit = 0;
  window_position = 0;
  sample_is_full = false;
  for(int i=0; i<progress_window; i++) sampled_fit[i] = 0;
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::report_now()
{
  if(report_interval >= running_time) return;

  Solution<T> which = current_solution;
  if(final_report){
    if(best_solution.get_fitness() < current_solution.get_fitness())
      which = best_solution;
  }

  ostream *out = Global<T>::out;
  //  SysTime sys;

  *out << "(";
  *out << "(solutions, " << monitor->get_solution_count() << ")\n";
  *out << "(time, " << monitor->exec_time() << ")\n";
  //  *out << "(sys_time, " << sys.get_time() << ")\n";
  *out << "(cost, " << which.get_cost() << ")\n";
#if 0
  *out << "(true_cost, " << which.get_true_cost() <<")\n";
  *out << "(true_correctness, " << which.get_true_correctness() <<")\n";
#endif
  if(ObjectiveFunction::correctness_is_defined())
    *out << "(correctness, " << which.get_correctness() << ")\n";

  *out << "(fitness, " << which.get_fitness() << ")\n";
  *out << "(size, " << which.get_size() << ")\n";

  if(ObjectiveFunction::correctness_is_defined()){
    *out << "(max_cf_delta, "<<Solution<T>::get_max_cf_delta()<<")\n";
    *out << "(min_cf_delta, "<<Solution<T>::get_min_cf_delta()<<")\n";
    *out << "(mean_cf_delta, "<<Solution<T>::get_mean_cf_delta()<<")\n";
    *out << "(mult_constant, "<<Solution<T>::get_mult_constant()<<")\n";
  }

  search_specific_display(out);
  *out << *(which.object);
  *out << ")\n";

  if(final_report){
    *out << "(true_cost, " << which.get_true_cost() <<")\n";
    *out << "(true_correctness, " << which.get_true_correctness() <<")\n";
  }
  out->flush();
}
//----------------------------------------------------------------------------------
template<class T>
bool BasicSearch<T>::check_progress_now()
{
  cp_count += 1.0;
  static double last_fit = -12345;
  double mean, variance;
  double fit = current_solution.get_fitness();

  if(fit == last_fit) {
    check_progress_interval *= 0.5;
    if(check_progress_interval < MinProgInterval) check_progress_interval = MinProgInterval;
  }
  else {
    check_progress_interval *= 2.0;
    if(check_progress_interval > MaxProgInterval) check_progress_interval = MaxProgInterval;
  }
  monitor->set_check_progress_interval(check_progress_interval);
  last_fit = fit;

  if(! sample_is_full){
    sampled_fit[window_position] = fit;
    window_position++;
    sum_fit += fit;
    sum_squares_fit += (fit*fit);
    mean = sum_fit/(double)window_position;
    variance = sum_squares_fit/(double)(window_position) - mean*mean;

    if(window_position == progress_window){
      window_position = 0;
      sample_is_full = true;
    }      
  }
  else {
    double old = sampled_fit[window_position];
    sampled_fit[window_position] = fit;
    window_position = (window_position + 1) % progress_window;
    sum_fit = sum_fit - old + fit;
    sum_squares_fit = sum_squares_fit - old*old + fit*fit;
    mean = sum_fit/(double)progress_window;
    variance = sum_squares_fit/(double)progress_window - mean*mean;
  }

#if 0 //debug
  for(int i=0;i<progress_window;i++) cout<<sampled_fit[i]<<" "; cout<<endl;
  cout<<"#check_progress_now(): "<<mean<<" "<<variance<<" "<<variance/(mean * mean)<<"  "<<variance_threshold<<endl;
#endif

  if(!sample_is_full && window_position<10) return true; //not enough sample
  if (variance>0 && variance/(mean * mean) > variance_threshold)
    {
      return true;  // there is still progress
    }

  // there is no more progress, predictively
#if 0 //debug
  cout << "#BasicSearch<T>::check_progress_now: no progress\n";
  cout.flush();
#endif

  return false;  
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::prepare_to_terminate()
{
#if 0 //debug
  std::cout << "# BasicSearch<T>::prepare_to_terminate()\n";
#endif
  //  search_exit();
  if(best_solution.object != NULL)
    Global<T>::update_best_solution(best_solution);
  set_final_report();
  report_now();

  //  clean_up();
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::gen_random_solution(Solution<T> * new_sol, int n)
{
  if(n==-1) n = solution_size;
  do {
    if(new_sol->gen_random_solution(n, monitor) == true) {
      search_stat = stat_cont;
      return;
    }
    search_stat = check_monitor_status();
  }while(search_stat == stat_cont);
  if(search_stat == stat_terminate) prepare_to_terminate();
}
//----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::gen_random_neighbor(Solution<T> * s, Solution<T> * new_sol)
{
  do{
    if(s->gen_random_neighbor(new_sol, monitor) == true){
      search_stat = stat_cont;
      return;
    }
    search_stat = check_monitor_status();
  }while(search_stat == stat_cont);
  if(search_stat == stat_terminate) prepare_to_terminate();
}

//----------------------------------------------------------------------------------

template<class T>
void BasicSearch<T>::restart_request()
{
  reset_window();
  monitor->reset_status(stat_check_progress);
  no_progress = false;
  search_stat = stat_cont;
}

//----------------------------------------------------------------------------------
template<class T>
Status BasicSearch<T>::check_monitor_status()
{
  MonitorStat stat = monitor->get_status();
  if(stat.test(stat_terminate)){
    return stat_terminate;    // must quit, there's no other way
  }

  if(stat.test(stat_report)){
    if(should_monitoring()) report_now();
    monitor->reset_status(stat_report);
  }
  
  if(stat.test(stat_check_progress)){
    if(should_monitoring()){
      if(no_progress) return stat_check_progress;
      
      if(check_progress_now() == false){
	no_progress = true;
	reset_window();
#if 0 //debug
	std::cout<<"# BasicSearch<T>::check_monitor_status(): no progress, stop this search\n";
#endif
	return stat_check_progress;
      }
    }
    monitor->reset_status(stat_check_progress);
  }

  return stat_cont;
}
//-----------------------------------------------------------------------------------
template<class T>
void BasicSearch<T>::strict_descending(double sample_constant, bool forced)
{
#if debug
  cout<<"\t#BasicSearch::strict_descending() "<<search_stat<<" "<<stat_check_progress<<"\n";
#endif

  if(forced){
    reset_window();
    monitor->reset_status(stat_check_progress);
    no_progress = false;
    //    check_progress_interval = MaxProgInterval;
  }

  double tries=0, max_tries;
  
  max_tries = sample_constant * current_solution.neighborhood_size() * log((double)current_solution.neighborhood_size());

  //  cout<<current_solution.get_fitness()<<endl;

  while(tries < max_tries){
    Solution<T> neighbor;
    gen_random_neighbor(&current_solution, &neighbor);

    if(search_stat != stat_cont) break;
    else if(current_solution.get_fitness() > neighbor.get_fitness()){
      tries = 0;
      current_solution = neighbor;
    }
    else{
      tries += 1.0;
    }
  }

  //  cout<<current_solution.get_fitness()<<endl;
}

//----------------------------------------------------------------------------

#endif
