/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef ENGINE_H
#define ENGINE_H

#include "basic_types.h"
#include "basic_search.h"

#include "Heuristics/hill_climb.cpp"
#include "Heuristics/simulated_anneal.cpp"
#include "Heuristics/sa_restart.cpp"
#include "Heuristics/greedy.cpp"
#include "Heuristics/smart_sa.cpp"
#include "Heuristics/g_smart_sa.cpp"
#include "Heuristics/cf_sa.cpp"
#include "Heuristics/rate_sa.cpp"
#include "Heuristics/dynamic_sa.cpp"
#include "Heuristics/g_cf_sa.cpp"

//#include "solution.h" #include "operator.h" #include "objective_function.h"

//----------------------------------------------------------------------------------------
// ENGINE
//----------------------------------------------------------------------------------------

/**
manages searching, singular and combined heuristics.
@author Vinhthuy Phan
@version 1.0
*/
template<class T>
class Engine {
public:
  ///
  BasicSearch<T> *the_search;
  ///
  HeuristicType heuristic_type;
  ///
  double running_time;
  ///
  double r_interval;
  ///
  double cp_interval;

  ///
  Engine(HeuristicType h, double report_interval, double check_progress_interval);

  ///
  ~Engine() ;
  
  ///
  bool run_once(HeuristicType h, double running_time, 
		double report_time, double check_time, 
		Solution<T> * init_sol,
		bool extra_flag);
  ///
  void run_loop();
  ///
  void hc_greedy_hc();
  ///
  void greedy_sa();
  ///
  void greedy_sa2();
  ///
  void greedy_sa3();
  ///
  void combined_search();
  ///
  void post_process();
};

// template class Engine<Permutation>;
// #include "basic_search.h"

//----------------------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------------------
template<class T>
Engine<T>::Engine(HeuristicType h,
		  double report_interval,
		  double check_progress_interval)
{
  heuristic_type = h;
  r_interval = report_interval;
  cp_interval = check_progress_interval;
  the_search = NULL;
}

//----------------------------------------------------------------------------------------
template<class T>
Engine<T>::~Engine()
{
  if(the_search!=NULL) delete the_search;
}


//----------------------------------------------------------------------------------------
template<class T>
void Engine<T>::post_process()
{
  if(the_search != NULL){
    the_search->set_final_report();
    the_search->report_now();
  }
  Global<T>::prepare_to_terminate();
}
//----------------------------------------------------------------------------------------
template<class T>
bool Engine<T>::run_once(HeuristicType h, double running_time, 
			 double report_time, double check_time, 
			 Solution<T> * init_sol,
			 bool smart_sa_flag)
{
  //  if(running_time < 0.5) return false;
  if(the_search!=NULL) delete the_search;

  switch(h){
  case h_hill_climb:
    the_search = new HillClimb<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_simulated_anneal:
    the_search = new SimulatedAnneal<T> (running_time, report_time, check_time, init_sol);
    break;
    //  case h_sa_restart:
    //    the_search = new SARestart<T> (running_time, report_time, check_time, init_sol);
    //    break;
  case h_smart_sa:
    the_search = new SmartSA<T> (running_time, report_time, check_time, init_sol, smart_sa_flag);
    break;
  case h_g_smart_sa:
    the_search = new GSmartSA<T> (running_time, report_time, check_time, init_sol, smart_sa_flag);
    break;
  case h_cf_sa:
    the_search = new CF_SA<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_g_cf_sa:
    the_search = new GCF_SA<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_rate_sa:
    the_search = new RateSA<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_dynamic_sa:
    the_search = new DynamicSA<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_greedy:
    the_search = new Greedy<T> (running_time, report_time, check_time, init_sol);
    break;
  case h_combined:
    combined_search();
    break;
  default:
    cout<<"invalid heuristic\n";
    exit(0);
  }

  if(the_search == NULL || the_search->is_enough_time() == false){
    //cout<<"not enough time: "<<the_search->overhead_time()<<"  "<<the_search->is_enough_time()<<endl;
    return false;
  }
  the_search->search();
  Global<T>::update_best_solution(the_search->best_solution);
  return true;
}

//----------------------------------------------------------------------------------------
template<class T>
void Engine<T>::run_loop()
{
#if debug
  cout <<"#Engine::run_loop\n";
#endif
  while(run_once(heuristic_type,
		 Global<T>::time_left(),
		 r_interval,
		 cp_interval,
		 NULL,
		 false))
    {
#if 0
      cout<<the_search->best_solution.get_fitness()<<endl;
#endif
    }

  //  the_search->make_valid();
  //  Global<T>::update_best_solution(the_search->best_solution);

  post_process();
}

//----------------------------------------------------------------------------------------
// the greedy heuristic chosen here used to be greedyK2 before it was
// chosen as the standard greedy.
//----------------------------------------------------------------------------------------
template<class T>
void Engine<T>::combined_search()
{
  /*
    combined strategy:

    - t= 0.05 of total time;
    - let t2 be the time HC finishes in t seconds.
    - run greedy in t2 units.
    - run cf_sa3La in t2 units.
    
    notation: > is better, >> much better (30%).

    * if HC > SA > G, HC first then SA.
    
    * if G >> HC and SA, run G.
    
    * if G and SA within 10% of each other, G first then SA.
    if greedy 2*t2 times and scale the new time proportionally to the performance.  The rest of the time is given to SA.
    
    otherwise, run SA.
  */
  //  double budget_fraction = 0.03, constant1 = 1.3, constant2 = 1.5;
  //  double budget_fraction = 0.01, constant1 = 1.3, constant2 = 1.5;
  //  double budget_fraction = 0.02, constant1 = 1.3, constant2 = 1.5;
  double budget_fraction = 0.04, constant1 = 1.3, constant2 = 1.5;
  double hc_fit, sa_fit, g_fit, time_budget = Global<T>::time_left() * budget_fraction;

  if(run_once(h_hill_climb, time_budget, r_interval, cp_interval, NULL, false) != false){
    hc_fit = the_search->best_solution.get_fitness();
    
    run_once(h_cf_sa, time_budget, r_interval, cp_interval, NULL, false);
    sa_fit = the_search->best_solution.get_fitness();

    run_once(h_greedy, time_budget, r_interval, cp_interval, NULL, false);
    g_fit = the_search->best_solution.get_fitness();
    
    if(hc_fit < sa_fit && hc_fit < g_fit){
      //cout<<"cb1 "<<hc_fit<<" "<<sa_fit<<" "<<g_fit<<endl;
      Global<T>::output_code = "cb1";
      if(run_once(h_hill_climb, Global<T>::time_left(), r_interval, cp_interval, NULL, false) == false){
	post_process();
	return;
      }      
    }
    else if(sa_fit > g_fit*constant2 && hc_fit > g_fit*constant2){
      //cout<<"cb2 "<<hc_fit<<" "<<sa_fit<<" "<<g_fit<<", time: "<<Global<T>::time_left()<<endl;
      Global<T>::output_code = "cb2";    
      while(run_once(h_greedy, Global<T>::time_left(), r_interval, cp_interval, NULL, false)){
	//	cout << the_search->best_solution.get_fitness()<<endl;
	;
      }

      //      cout << the_search->best_solution.get_fitness()<<endl;
					   
      post_process();
      return;
    }
    else if(sa_fit > g_fit*constant1 && hc_fit > g_fit*constant1){
      //cout<<"cb3 "<<hc_fit<<" "<<sa_fit<<" "<<g_fit<<", time: "<<Global<T>::time_left()<<endl;
      Global<T>::output_code = "cb3";    
      if(run_once(h_greedy, Global<T>::time_left(), r_interval, cp_interval, NULL, false) == false){
	post_process();
	return;
      }
    }
    else if((g_fit > sa_fit && (g_fit-sa_fit)/g_fit <= 0.1) ||
	    (sa_fit > g_fit && (sa_fit-g_fit)/sa_fit <= 0.1)){
      //cout<<"cb4 "<<hc_fit<<" "<<sa_fit<<" "<<g_fit<<endl;
      Global<T>::output_code = "cb4";
      run_once(h_greedy, 2.0*time_budget, r_interval, cp_interval, NULL, false);
      double g_split = 0.2, p = the_search->best_solution.get_fitness()/g_fit;
      if(p < 1.0 && 1.0-p > g_split){
	g_split = 1.0-p;
      }
      if(run_once(h_greedy, g_split * Global<T>::time_left(), r_interval, cp_interval, NULL, false) == false){
	post_process();
	return;
      }
      Solution<T> init_sol = the_search->best_solution;
      if(run_once(h_cf_sa, Global<T>::time_left(), r_interval, cp_interval, &init_sol, true) == false){
	post_process();
	return;
      }
    }
  }

  // by default or all have finished, run cf_sa
  if(Global<T>::time_left() > 0){
    //    cout<<"here\n";
    run_once(h_cf_sa, Global<T>::time_left(), r_interval, cp_interval, NULL, false);
  }
  post_process();
}
//----------------------------------------------------------------------------------------
template<class T>
void Engine<T>::hc_greedy_hc()
{
#if debug
  cout <<"#Engine::hc_greedy_hc()\n";
#endif  

  double tBegin, tEnd, hc_time;

  // -- Hill Climb
  tBegin = Global<T>::time_left();
  run_once(h_hill_climb, Global<T>::time_left(), r_interval, 0.1, NULL, false);
  tEnd = Global<T>::time_left();
  hc_time = tBegin - tEnd;

  while((running_time = Global<T>::time_left()) > 0){
    //-------------------------------------------
    // -- Greedy
    if(run_once(h_greedy, running_time - hc_time, r_interval, cp_interval, NULL, false) == false)
      break;
  
    //-------------------------------------------
    // -- Hill Climb
#if debug
    cout<<"#run hc for "<<hc_time<<" seconds on best solution ("<<the_search->best_solution.get_fitness()<<")\n";
#endif
    Solution<T> init_sol = the_search->best_solution;
    if(run_once(h_hill_climb, hc_time, r_interval, cp_interval, &init_sol, false) == false)
      break;
  }
    
  post_process();
}


//----------------------------------------------------------------------------------------

template<class T>
void Engine<T>::greedy_sa()
{
#if debug
  cout <<"#Engine::greedy_sa()\n";
#endif  

  double min_greedy_fit = 999999999, min_sa_fit = 999999999;

  while((running_time = Global<T>::time_left()) > 0){
    //-------------------------------------------
    // -- Greedy + HC
    if(run_once(h_greedy, running_time * 0.5, r_interval, cp_interval, NULL, false) == false)
      break;

    Solution<T> init_sol = the_search->best_solution;

    if(run_once(h_hill_climb, Global<T>::time_left(), r_interval, 0.1, &init_sol, false) == false)
      break;

    if(the_search->current_solution.get_fitness() < min_greedy_fit)
      min_greedy_fit = the_search->current_solution.get_fitness();
    
    //-------------------------------------------
    // -- "Smart" SA
    init_sol = the_search->best_solution;
    if(run_once(h_smart_sa, Global<T>::time_left(), r_interval, cp_interval, &init_sol, true) == false)
      break;
    if(the_search->current_solution.get_fitness() < min_sa_fit)
      min_sa_fit = the_search->current_solution.get_fitness();
  }
  if (min_greedy_fit < min_sa_fit) Global<T>::output_code = "GH";
  else Global<T>::output_code = "SA";

  post_process();
}

//----------------------------------------------------------------------------------------

template<class T>
void Engine<T>::greedy_sa3()
{
#if debug
  cout <<"#Engine::greedy_sa3()\n";
#endif  

  double min_greedy_fit = 999999999, min_sa_fit = 999999999;

  while((running_time = Global<T>::time_left()) > 0){
    //-------------------------------------------
    // -- Greedy + HC
    if(run_once(h_greedy, running_time * 0.5, r_interval, cp_interval, NULL, false) == false)
      break;

    Solution<T> init_sol = the_search->best_solution;

    if(run_once(h_hill_climb, Global<T>::time_left(), r_interval, 0.1, &init_sol, false) == false)
      break;

    if(the_search->current_solution.get_fitness() < min_greedy_fit)
      min_greedy_fit = the_search->current_solution.get_fitness();
    
    //-------------------------------------------
    // -- "Smart" GSA
    init_sol = the_search->best_solution;
    if(run_once(h_g_smart_sa, Global<T>::time_left(), r_interval, cp_interval, &init_sol, true) == false)
      break;

    if(the_search->current_solution.get_fitness() < min_sa_fit)
      min_sa_fit = the_search->current_solution.get_fitness();
  }
  if (min_greedy_fit < min_sa_fit) Global<T>::output_code = "GH";
  else Global<T>::output_code = "SA";

  post_process();
}

//----------------------------------------------------------------------------------------

template<class T>
void Engine<T>::greedy_sa2()
{
#if debug
  cout <<"#Engine::greedy_sa()\n";
#endif  

  double min_greedy_fit = 999999999, min_sa_fit = 999999999;
  double r = 0.05 * Global<T>::time_left(), hc_result=0, g_result=0;
  if(r<1) r = 1;
  if(run_once(h_hill_climb, r, r_interval, 0.1, NULL, false) == true){
    hc_result = the_search->best_solution.get_fitness();
    if(run_once(h_greedy, r, r_interval, 0.1, NULL, false) == true){
      g_result = the_search->best_solution.get_fitness();
    }
  }

  while((running_time = Global<T>::time_left()) > 0){
    double g_time = Global<T>::time_left() * hc_result/(hc_result+g_result);
    //    double sa_time = Global<T>::time_left() * g_result/(hc_result+g_result);

#if debug
    cout<<"greedy_sa2(): r = "<<r<<endl;
    cout<<"greedy_sa2(): assigning "<<g_time<<" units to G and about "<<Global<T>::time_left()-g_time<<" units to SA\n";
#endif

    //-------------------------------------------
    // -- Greedy + HC
    if(run_once(h_greedy, g_time, r_interval, cp_interval, NULL, false) == false)
      break;
    Solution<T> init_sol = the_search->best_solution;
    if(run_once(h_hill_climb, Global<T>::time_left(), r_interval, 0.1, &init_sol, false) == false)
      break;
    if(the_search->current_solution.get_fitness() < min_greedy_fit)
      min_greedy_fit = the_search->current_solution.get_fitness();

    
    //-------------------------------------------
    // -- "Smart" SA
    init_sol = the_search->best_solution;
    if(run_once(h_smart_sa, Global<T>::time_left(), r_interval, cp_interval, &init_sol, true) == false)
      break;

    if(the_search->current_solution.get_fitness() < min_sa_fit)
      min_sa_fit = the_search->current_solution.get_fitness();
  }
  
  if (min_greedy_fit < min_sa_fit) Global<T>::output_code = "GH";
  else Global<T>::output_code = "SA";

  post_process();
}

//----------------------------------------------------------------------------------------


#endif
