/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef GLOBAL_H
#define GLOBAL_H

#include "solution.h"
#include "operator.h"
#include "objective_function.h"

// FORWARD DECLARATIONS for global use
template <class T> class Solution;

//----------------------------------------------------------------------------------------
// GLOBAL
//----------------------------------------------------------------------------------------
/**
manages global activities and computations (such as creation times).
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class Global {
private:
  ///
  static Monitor *monitor;

  ///
  static double search_overhead_time;

  /// --- these are NOT computed in constructor for efficient reason
  static double solution_creation_time;
  ///
  static double neighbor_creation_time;

  ///
  static double max_cost;
  ///
  static double min_cost;
  ///
  static double max_correctness;
  ///
  static double min_correctness;
  ///
  static double global_running_time;

public:
  ///
  static Solution<T> best_solution;
  ///
  static char *input_file;
  ///
  static ostream *out;
  ///
  static NeighborOp<T> * neighborhood_op;
  ///
  static char * output_code;
  ///
  Global();
  ///
  static void initialize(double running_time, NeighborOpType op_type, ostream * output, char *input);
  ///
  static void clean_up();
  ///
  ~Global();
  ///
  static void compute_search_space_statistics();
  ///
  static void prepare_to_terminate();
  ///
  static void update_best_solution(Solution<T> &s);
  ///
  static double time_left();
  ///
  static double get_search_overhead_time() { return search_overhead_time; }
  ///
  static double get_solution_creation_time() { return solution_creation_time; }
  ///
  static double get_neighbor_creation_time() { return neighbor_creation_time; }
  ///
  static double get_max_cost() { return max_cost; }
  ///
  static double get_min_cost() { return min_cost; }
  ///
  static double get_max_correctness() { return max_correctness; }
  ///
  static double get_min_correctness() { return min_correctness; }
  ///
  static double get_global_running_time();
  ///
  static void start_global_clock();
  ///
  static void reset_global_clock();
};

template <class T> char * Global<T>::output_code = NULL;

template <class T> char * Global<T>::input_file = NULL;

template <class T> ostream * Global<T>::out = NULL;

template <class T> Solution<T> Global<T>::best_solution;

template <class T> NeighborOp<T> * Global<T>::neighborhood_op = NULL;

template <class T> double Global<T>::search_overhead_time = 0;

template <class T> double Global<T>::solution_creation_time = 0;

template <class T> double Global<T>::neighbor_creation_time = 0;

template <class T> Monitor * Global<T>::monitor = NULL;

template <class T> double Global<T>::max_cost = 0;

template <class T> double Global<T>::min_cost = 0;

template <class T> double Global<T>::max_correctness = 0;

template <class T> double Global<T>::min_correctness = 0;

template <class T> double Global<T>::global_running_time = 0;


//#include "permutation.h"
//template class Global<Permutation>;
//#include "global.h"

//----------------------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------------------
template<class T>
Global<T>::Global(){  ;  }
//----------------------------------------------------------------------------------------
template<class T>
Global<T>::~Global()
{
  if(monitor != NULL) delete monitor;
}
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::update_best_solution(Solution<T> & s)
{
  if(best_solution.object == NULL ||
     (best_solution.get_fitness() > s.get_fitness() && !s.is_partial() && s.get_true_correctness() == 0)){
    //    best_solution.copy(s);
    //    best_solution.consolidate();
    //    cout<<"update best solution: "<<best_solution.object<<" "<<s.is_mutated()<<" "<<s.get_correctness()<<" "
    //	<< s.get_fitness()<<"  "<<s.get_num_elements()<<" "<<ObjectiveFunction::get_size()<<endl;


    best_solution = s;
  }
}
//----------------------------------------------------------------------------------------
template<class T>
double Global<T>::time_left()
{
  return monitor->time_left();
}
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::prepare_to_terminate()
{
#if debug
  std::cout << "# Global::prepare_to_terminate()\n";

  if(best_solution.object != NULL){
    *out << "(";
    *out << "(time, " << monitor->exec_time() << ")\n";
    *out << "(cost, " << best_solution.get_cost() << ")\n";
    if(ObjectiveFunction::correctness_is_defined())
      *out << "(correctness, " << best_solution.get_correctness() << ")\n";
    *out << "(fitness, " << best_solution.get_fitness() << ")\n";
    *out << "(true_cost, " << best_solution.get_true_cost() << ")\n";
    *out << "(true_correctness, " << best_solution.get_true_correctness() << ")\n";

    //    *out << "***** is mutated = "<<best_solution->is_mutated() << endl;
    //    *out << "(content, " ;    best_solution->display(*out);    *out << ")\n";
  }
  *out << "(solutions, " << monitor->get_solution_count() << ")\n";
  *out << "EOF" << endl;
  out->flush();

  //  Global<T>::clean_up();

  *out <<"# "<<T::constructor_count1+T::constructor_count2+T::constructor_count3+T::constructor_count4
       <<" ("<<T::constructor_count1<<" + "<<T::constructor_count2
       <<" + "<<T::constructor_count3<<" + "<<T::constructor_count4<<") constructed, "
       <<T::destructor_count<<" destroyed\n";
  
  if(ObjectiveFunction::correctness_is_defined()){
    *out << "# Solution::max_cf_delta "<<Solution<T>::get_min_cf_delta()<<endl;
    *out << "# Solution::min_cf_delta "<<Solution<T>::get_min_cf_delta()<<endl;
  }
#else
  if(best_solution.object == NULL){
    *out <<input_file<<" ";
    *out << "best_solution is NULL\n";
  }
  else {
    if(best_solution.is_mutated()){
      //      cout<<"best solution is mutated\n";
      best_solution.consolidate();
    }

    *out <<input_file<<" ";
    *out <<monitor->get_solution_count()<<" "
	 <<best_solution.get_fitness()<<" "
	 <<best_solution.get_cost()<<" "<<best_solution.get_true_cost()<<" "
	 <<best_solution.get_correctness()<<" "<<best_solution.get_true_correctness()<<" "
	 <<best_solution.get_num_elements()<<" "<<ObjectiveFunction::get_size();
	if(output_code == NULL)
		*out << " (NULL)\n";
	else *out << " "<<output_code<<endl;
	
    //    *out << "(content, " ;    best_solution.display(*out);    *out << ")\n";
  }
#endif
  //exit(0);
}
//----------------------------------------------------------------------------------------
// compute (1) times it takes to create a new solution, neighbor;
//         (2) estimated maximal, minimal costs and correctness's.
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::compute_search_space_statistics()
{
#if debug
  std::cout<<"#Global::compute_search_space_statistics() search_overhead_time "<<search_overhead_time<<endl;
#endif

  // -- compute time to create a random solution and a neighbor
  Solution<T> new_sol, new_neighbor;
  int counter;
  double running_time = search_overhead_time * 0.5;
  Date now, start, time_span(running_time);
  for(counter=0; (now-start) < time_span; counter++){
    if(new_sol.gen_random_solution(ObjectiveFunction::get_size(), monitor) == false)
      return;

    if(counter == 0){
      max_cost = min_cost = new_sol.get_cost();
      max_correctness = min_correctness = new_sol.get_correctness();
    }
    else {
      if(max_cost < new_sol.get_cost()) max_cost = new_sol.get_cost();
      if(min_cost > new_sol.get_cost()) min_cost = new_sol.get_cost();
      if(ObjectiveFunction::correctness_is_defined()){
	if(max_correctness < new_sol.get_correctness()) max_correctness = new_sol.get_correctness();
	if(min_correctness > new_sol.get_correctness()) min_correctness = new_sol.get_correctness();
      }
    }
    now = Date();
  }
  now = Date();
  now = now - start;
  solution_creation_time = now.get_time() / (double)counter;

  start = Date();
  if(new_sol.gen_random_solution(ObjectiveFunction::get_size(), monitor) == false)
    return;

  for(counter = 0; (now-start) < time_span; counter++){
    if(new_sol.gen_random_neighbor(&new_neighbor, monitor) == false)
      return;

    if(max_cost < new_sol.get_cost()) max_cost = new_sol.get_cost();
    if(min_cost > new_sol.get_cost()) min_cost = new_sol.get_cost();
    if(ObjectiveFunction::correctness_is_defined()){
      if(max_correctness < new_sol.get_correctness()) max_correctness = new_sol.get_correctness();
      if(min_correctness > new_sol.get_correctness()) min_correctness = new_sol.get_correctness();
    }
    now = Date();
  }
  now = Date();
  now = now - start;
  neighbor_creation_time = now.get_time() / counter;
}
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::clean_up()
{
  if(monitor != NULL) delete monitor;
}
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::initialize(double running_time, NeighborOpType op_type, ostream * output, char *input)
{
  input_file = input;
  out = output;
  global_running_time = running_time;
  search_overhead_time = max(0.1, 0.001 * running_time);
  start_global_clock();

  switch(op_type){
  case Swap:
    neighborhood_op = new SwapOp<T>();
    break;
  default:
    std::cerr<<"unknown operator\n";
    exit(0);
  }
    
  compute_search_space_statistics();
#if debug
  cout<<"#Global::initialize() min/max cost min/max correctness: "
      <<min_cost <<" "<<max_cost<<" "<< min_correctness<<" "<<max_correctness<<endl;
#endif
}

//----------------------------------------------------------------------------------------
template<class T>
double Global<T>::get_global_running_time()
{
  return global_running_time;
}
//----------------------------------------------------------------------------------------

template<class T>
void Global<T>::start_global_clock()
{
  monitor = new Monitor(non_search_mode, global_running_time);
}
//----------------------------------------------------------------------------------------
template<class T>
void Global<T>::reset_global_clock()
{
  if(monitor != NULL) delete monitor;
  monitor = new Monitor(non_search_mode, global_running_time);
}
//----------------------------------------------------------------------------------------


#endif
