/*
Copyright (C) 2003 copyrighted by SUNY at Stony Brook.

DISCROPT is created by Vinhthuy Phan, Pavel Sumazin, and Steven Skiena.
The authors grant permision to use and distribute WITHOUT ANY WARRANTY
this software for non-commercial purposes, provided that this copyright 
notice is clearly shown.  Other uses are prohibited without prior written
consent.  These terms are applied to any modification of this software 
as well.  See http://www.cs.sunysb.edu/~discropt.

*/
#ifndef SOLUTION_H
#define SOLUTION_H

#include <iostream>
using std::cout;
using std::endl;

//#include <ostream>
//using std::ostream;

#include "basic_types.h"
#include "monitor.h"
#include "global.h"
#include "objective_function.h"
#include "problem_defs.h"

#define TEST_DELTA_COST  0

/// FORWARD DECLARATIONS for global use
template <class T> class Global;

//--------------------------------------------------------------------
//----------------------------------------------------------------------------------------
// SOLUTION
//----------------------------------------------------------------------------------------

/**
keeps track of solution object and its fitness with "just-in-time" computations
of fitness.
@author Vinhthuy Phan
@version 1.0
*/
template <class T>
class Solution {
private:
  
  /** At any given time, a solution stores correct values of each of these,
      regardless whether or not is mutated */
  double cost;
  ///
  double correctness;
  ///
  mutation_element mut_element;
  ///
  void commit();

  /// --- New implementation of cost/correctness combination
  static double max_cf_delta;
  ///
  static double min_cf_delta;
  ///
  static double mean_cf_delta;
  ///
  static double mult_constant;
  ///
  static double cf_count;

public:
  ///
  T *object;
  ///
  void commit_extension();
  ///
  Solution(T & object);
  ///
  Solution();
  ///
  Solution(const Solution<T> & sol);
  ///
  ~Solution();
  ///
  void extend(int pos, int ele);

  /// Generation of new solutions
  bool gen_random_neighbor(Solution<T> * new_sol, Monitor * monitor);
  ///
  bool gen_random_solution(int n, Monitor * monitor);
  ///
  void gen_extension(Solution<T> *new_sol, int pos, int ele);

  /// ---- Assignment, no new solution generated
  Solution<T> & operator=(Solution<T> &sol);
  ///
  bool is_mutated() const;
  ///
  mutation_element get_mutation_element() const;
  ///
  double get_fitness() const;
  ///
  double get_cost() const;
  ///
  double get_correctness() const;
  ///
  double neighborhood_size() const;
  ///
  void consolidate(bool totally = false);

  /// --- supposed to be used only at exit
  double get_true_cost() const;
  ///
  double get_true_correctness() const;
  ///
  std::ostream & display(std::ostream & out);
  ///
  int get_refCount () const { return object->get_refCount(); }
  ///
  static double get_max_cf_delta() { return max_cf_delta; }
  ///
  static double get_min_cf_delta() { return min_cf_delta; }
  ///
  static double get_mean_cf_delta() { return mean_cf_delta; }
  ///
  static double get_mult_constant() { return mult_constant; }
  ///
  static void add_to_max_cf_delta(double dk)
  {
    max_cf_delta += dk;
  }
  ///
  static void add_to_min_cf_delta(double dk)
  {
    min_cf_delta += dk;
  }
  ///
  static void add_to_mean_cf_delta(double dk)
  {
    mean_cf_delta += dk;
  }
  ///
  static void add_to_mult_constant(double dk)
  {
    //    cout<<"mult_constant = "<<mult_constant<<endl;
    mult_constant += dk;
  }

  /// ---- calls to object
  T * get_object() { return object; }
  ///
  int get_num_elements() const { return object->get_num_elements(); }
  ///
  int get_size() const { return object->get_size(); }
  ///
  bool is_partial() const { 
    return (object->get_num_elements() < ObjectiveFunction::get_size());
  }

  //  int first_position() const { return object->first_position(); }
  //  int last_position() const { return object->last_position(); }
  ///
  static bool is_ordered() { return T::is_ordered(); }
};

//--------------------------------------------------------------------
// Initialization of best_delta_value
//--------------------------------------------------------------------


template <class T> double Solution<T>::max_cf_delta = 0;

template <class T> double Solution<T>::min_cf_delta = 0;

template <class T> double Solution<T>::mean_cf_delta = 0;

template <class T> double Solution<T>::cf_count = 0;

template <class T> double Solution<T>::mult_constant = 0;

//--------------------------------------------------------------------
//--------------------------------------------------------------------
// IMPLEMENTATION
//--------------------------------------------------------------------
// Class T (permutation, subset, ...) must provide the basic constructors
// such as random (given size), copy, etc.
// ObjectiveFunction must provide static functions: cost, correctness
// appropriately defined.
//--------------------------------------------------------------------
//--------------------------------------------------------------------
// default and copy constructor create fresh (unattached) *object's.
//--------------------------------------------------------------------
//--------------------------------------------------------------------
#if 0
template<class T> 
void Solution<T>::copy(const Solution<T> & sol)
{
  this->object = new T(*(sol.object));
  this->object->incre_refCount();
  cost = ObjectiveFunction::cost(*object);
  if(ObjectiveFunction::correctness_is_defined())
    correctness = ObjectiveFunction::correctness(*object);
  else correctness = 0;
  mut_element.operation = NoOp;
}
#endif

//--------------------------------------------------------------------
template<class T> 
Solution<T>::Solution(T & object)
{
  this->object = new T(object);
  this->object->incre_refCount();
  cost = ObjectiveFunction::cost(object);
  if(ObjectiveFunction::correctness_is_defined())
    correctness = ObjectiveFunction::correctness(object);
  else correctness = 0;
  mut_element.operation = NoOp;
}
//--------------------------------------------------------------------
template<class T> 
Solution<T>::Solution()
{
  object = NULL;
  cost = correctness = 0;
  mut_element.operation = NoOp;
}
//--------------------------------------------------------------------
template< class T >
Solution<T>::Solution(const Solution<T> & sol)
{
  // worse:  object = new T(*(sol.object));
  object = sol.object;
  object->incre_refCount();
  cost = sol.cost;
  correctness = sol.correctness;
  mut_element = sol.mut_element;
}
//--------------------------------------------------------------------

template< class T >
void Solution<T>::extend(int pos, int ele)
{
  //  cout<<"extend: "<<pos<<" "<<ele<<endl;

  mutation_element mut;
  mut.first = pos;
  mut.second = ele;
  cost +=  ObjectiveFunction::extend_cost(*object, mut);
  if(ObjectiveFunction::correctness_is_defined()){
    correctness += ObjectiveFunction::extend_correctness(*object, mut);
  }

  T * tmp_object = new T(*object, pos, ele);
  if(tmp_object == NULL){
    cout<<"extend: out of mem\n";
    exit(0);
  }
  tmp_object->incre_refCount();
  this->object->decre_refCount();
  if(this->object->get_refCount() == 0) delete this->object;
  this->object = tmp_object;
  //  cout<<"extend: "<<object<<" "<<object->get_refCount()<<endl;
}

//--------------------------------------------------------------------
template< class T >
void Solution<T>::gen_extension(Solution<T> *new_sol, int pos, int ele)
{
  if(is_mutated()) commit_extension();

  new_sol->mut_element.first = pos;
  new_sol->mut_element.second = ele;
  new_sol->mut_element.operation = Global<T>::neighborhood_op->get_type();

  if(new_sol->object != NULL){
    new_sol->object->decre_refCount();
    if(new_sol->object->get_refCount() == 0) delete new_sol->object;
  }

  new_sol->object = this->object;
  new_sol->object->incre_refCount();

  new_sol->cost = cost + ObjectiveFunction::extend_cost(*object, new_sol->mut_element);
  if(ObjectiveFunction::correctness_is_defined())
    new_sol->correctness = correctness + ObjectiveFunction::extend_correctness(*object, new_sol->mut_element);
  else new_sol->correctness = 0;
}
//--------------------------------------------------------------------
template< class T >
bool Solution<T>::gen_random_solution(int n, Monitor *monitor)
{
  if(monitor!=NULL && monitor->block()) return false;

  if(object != NULL){
    object->decre_refCount();
    if(object->get_refCount() == 0) delete object;
  }
  object = new T(n);
  if(object == NULL){    cout<<"gen_random_solution: out of mem\n";    exit(0);  }
  object->incre_refCount();
  mut_element.operation = NoOp;
  cost = ObjectiveFunction::cost(*object);
  if(ObjectiveFunction::correctness_is_defined())
    correctness = ObjectiveFunction::correctness(*object);

  return true;
}
//--------------------------------------------------------------------
// *this is mutated (specified by mut_el) from sol.
//
// the mutation constructor provides the mean by which a search
// traverses from one solution to another.
// attach *this to sol->solution, which is detached if it's already
// mutated.
//--------------------------------------------------------------------
template< class T >
bool Solution<T>::gen_random_neighbor(Solution<T> * new_sol, Monitor *monitor)
{
  if(monitor!=NULL && monitor->block()) return false;

  if(new_sol->object != NULL){
    new_sol->object->decre_refCount();
    if(new_sol->object->get_refCount() == 0) delete new_sol->object;
  }

  if(is_mutated()){
    if(object->get_refCount() == 1) commit();
    else if(object->get_refCount() > 1){
      object->decre_refCount();
      object = new T(*object);
      if(object == NULL){ cout<<"gen_random_neighbor: out of mem\n";	exit(0);      }
      object->incre_refCount();
      commit();
    }
    else {
      cout << "refCount < 1, cannot mutate an empty solution\n";    exit(0);
    }
  }


  new_sol->object = object;
  new_sol->object->incre_refCount();
  Global<T>::neighborhood_op->gen_neighbor(*object, &(new_sol->mut_element));


  double d_correctness, d_cost = ObjectiveFunction::delta_cost(*object, new_sol->mut_element);
  new_sol->cost = cost + d_cost;


  if(ObjectiveFunction::correctness_is_defined()){
    d_correctness = ObjectiveFunction::delta_correctness(*object, new_sol->mut_element);
    new_sol->correctness = correctness + d_correctness;

    //-------- update best_delta_value -------//
    if(d_correctness != 0){
      double d_abs = d_cost/d_correctness;

      cf_count += 1.0;
      if(d_abs < 0) d_abs *= -1;
      if(max_cf_delta==0 || d_abs > max_cf_delta) max_cf_delta = d_abs;
      if(min_cf_delta==0 || d_abs < min_cf_delta) min_cf_delta = d_abs;
      if(cf_count <= 1.0) mean_cf_delta = d_abs;
      else mean_cf_delta = (mean_cf_delta * (cf_count-1.0) + d_abs)/cf_count;
      mult_constant = max_cf_delta + 1.0;
    }
  }
  else new_sol->correctness = 0;

#if TEST_DELTA_COST
  //----- to be kept. testing for correctness of delta cost transfer and maintainance
  Solution<T> * tmp_sol = new Solution<T>;
  tmp_sol->object = new T(*(new_sol->object));
  tmp_sol->mut_element = new_sol->mut_element;
  tmp_sol->commit();

  cout <<"mut (first, second): "<<(tmp_sol->mut_element).first<<" "
       <<(tmp_sol->mut_element).second<<endl;
  cout <<"real cost (old, new): ("<<ObjectiveFunction::cost(*object)<<", "
       <<ObjectiveFunction::cost(*(tmp_sol->object))<<")    "
       <<"new_sol cost: "<<new_sol->cost<<"  "
       <<" = "<< cost << " + " << d_cost<<endl;
  if(ObjectiveFunction::cost(*(tmp_sol->object)) != new_sol->cost){
    cout<<"cost in disagreement**********\n\n";
    cout << "old object "<<*object << endl;
    cout << "new object "<<*(tmp_sol->object) << endl;
    exit(0);
  }

  if(ObjectiveFunction::correctness_is_defined()){
    cout <<"real correctness (old, new): ("<<ObjectiveFunction::correctness(*object)<<", "
	 <<ObjectiveFunction::correctness(*(tmp_sol->object))<<")    "
	 <<"new_sol correctness: "<<new_sol->correctness<<"  "
	 <<" = "<< correctness << " + " << d_correctness<<endl;


    if(ObjectiveFunction::correctness(*(tmp_sol->object)) != new_sol->correctness){
      cout<<"correctness in disagreement***********\n\n";
      cout << "old object "<<*object << endl;
      cout << "new object "<<*(tmp_sol->object) << endl;
      exit(0);
    }
  }
#endif

  return true;
}
//--------------------------------------------------------------------
template< class T >
Solution<T>::~Solution()
{
  if(this->object != NULL){
    this->object->decre_refCount();
    cout.flush();
    if(this->object->get_refCount() == 0){
      delete this->object;
    }
    this->object = NULL;
  }
}
//--------------------------------------------------------------------
// detach *this from current *object, and attatch it to sol's
//--------------------------------------------------------------------
template< class T >
Solution<T>  & Solution<T>::operator=(Solution<T> & sol)
{
  if(&sol != this){
    if(sol.object != object){
      if(object != NULL){
	object->decre_refCount();
	if(object->get_refCount() == 0){
	  delete object;
	}
      }
      object = sol.object;
      if(object != NULL) object->incre_refCount();
    }
    cost = sol.cost;
    correctness = sol.correctness;
    mut_element = sol.mut_element;
  }
  return *this;
}

//--------------------------------------------------------------------
template<class T>
void Solution<T>::commit()
{
  Global<T>::neighborhood_op->commit(object, mut_element);
  mut_element.operation = NoOp;
}

//--------------------------------------------------------------------
template<class T>
void Solution<T>::commit_extension()
{
  T * tmp_object = new T(*object, mut_element.first, mut_element.second);
  if(tmp_object == NULL){
    cout<<"commit_extension: out of mem\n";
    exit(0);
  }
  tmp_object->incre_refCount();
  object->decre_refCount();
  if(object->get_refCount() == 0) delete object;
  object = tmp_object;
}

//--------------------------------------------------------------------
// totally == true means consolidate (commit mutation) and compute cost
template<class T>
void Solution<T>::consolidate(bool totally)
{
  if(is_mutated()){
    if(object->get_refCount() == 1){
      commit();
      //      cout<<"commit\n";
    }
    else if(object->get_refCount() > 1){
      object->decre_refCount();
      object = new T(*object);
      if(object == NULL){
	cout<<"consolidate: out of mem\n";
	exit(0);
      }
      object->incre_refCount();
      commit();
      //      cout<<"commit > 1\n";
    }
    else {
      cerr << "refCount < 1, cannot mutate an empty solution\n";    exit(0);
    }
  }

  if(totally){
    cost = ObjectiveFunction::cost(*object);
    if(ObjectiveFunction::correctness_is_defined())
      correctness = ObjectiveFunction::correctness(*object);
  }
}  
//--------------------------------------------------------------------
template<class T>
bool Solution<T>::is_mutated() const 
{
  return (mut_element.operation != NoOp);
}
//--------------------------------------------------------------------
template<class T>
mutation_element Solution<T>::get_mutation_element() const
{ 
  return mut_element;
}
//--------------------------------------------------------------------
template<class T>
double Solution<T>::get_fitness() const
{
  if(ObjectiveFunction::correctness_is_defined()) 
    return cost + mult_constant * correctness;
  return cost;
}
//--------------------------------------------------------------------
#if 0
template<class T>
double Solution<T>::get_fitness(T & it) 
{
  double cost, correctness;
  cost = ObjectiveFunction::cost(it);
  if(ObjectiveFunction::correctness_is_defined()) {
    correctness = ObjectiveFunction::correctness(it);
    return cost + correctness;
  }
  return cost;
}
#endif
//--------------------------------------------------------------------
template<class T>
double Solution<T>::get_true_cost() const
{
  return ObjectiveFunction::true_cost(*object);
}
//--------------------------------------------------------------------
template<class T>
double Solution<T>::get_true_correctness() const
{
  return ObjectiveFunction::true_correctness(*object);
}
//--------------------------------------------------------------------
template<class T>
double Solution<T>::get_cost() const
{
  return cost;
}
//--------------------------------------------------------------------
template<class T>
double Solution<T>::get_correctness() const 
{
  return correctness;
}
//--------------------------------------------------------------------

template<class T>
std::ostream & Solution<T>::display(std::ostream & out)
{
  out << *object;
  return out;
}

//--------------------------------------------------------------------
template<class T>
double Solution<T>::neighborhood_size() const
{
  return Global<T>::neighborhood_op->neighborhood_size(*object);
}
//--------------------------------------------------------------------

#endif
